"""
Generic NONMEM code record class.

"""

import re

import lark
import sympy
import sympy.printing.codeprinter
import sympy.printing.fortran
from sympy import Piecewise
from sympy.printing.str import StrPrinter

import pharmpy.symbols as symbols
from pharmpy.data_structures import OrderedSet
from pharmpy.parse_utils.generic import AttrToken, NoSuchRuleException
from pharmpy.plugins.nonmem.records.parsers import CodeRecordParser
from pharmpy.statements import Assignment, ModelStatements

from .record import Record


class MyPrinter(StrPrinter):
    def _print_Add(self, expr):
        args = expr.args
        new = []
        for arg in args:
            new.append(self._print(arg))
        return super()._print_Add(sympy.Add(*args, evaluate=False), order='none')


class NMTranPrinter(sympy.printing.fortran.FCodePrinter):
    # Differences from FCodePrinter in sympy
    # 1. Upper case
    # 2. Use Fortran 77 names for relationals
    # 3. Use default kind for reals (which will be translated to double kind by NMTRAN)
    # All these could be submitted as options to the sympy printer
    _relationals = {
        '<=': '.LE.',
        '>=': '.GE.',
        '<': '.LT.',
        '>': '.GT.',
        '!=': '.NE.',
        '==': '.EQ.',
    }
    _operators = {
        'and': '.AND.',
        'or': '.OR.',
        'xor': '.NEQV.',
        'equivalent': '.EQV.',
        'not': '.NOT. ',
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._settings["standard"] = 95

    def _print_Float(self, expr):
        printed = sympy.printing.codeprinter.CodePrinter._print_Float(self, expr)
        return printed


def nmtran_assignment_string(assignment, defined_symbols, rvs, trans):
    if isinstance(assignment.expression, Piecewise):
        statement_str = _translate_sympy_piecewise(assignment, defined_symbols)
    elif re.search('sign', str(assignment.expression)):  # FIXME: Don't use re here
        statement_str = _translate_sympy_sign(assignment)
    else:
        statement_str = _print_custom(assignment, rvs, trans)
    return statement_str


def _translate_sympy_piecewise(statement, defined_symbols):
    expression = statement.expression.args
    symbol = statement.symbol
    # Did we (possibly) add the default in the piecewise with 0 or symbol?
    has_added_else = expression[-1][1] is sympy.true and (
        expression[-1][0] == symbol or (expression[-1][0] == 0 and symbol not in defined_symbols)
    )
    if has_added_else:
        expression = expression[0:-1]
    has_else = expression[-1][1] is sympy.true

    expressions, _ = zip(*expression)

    if len(expression) == 1:
        value = expression[0][0]
        condition = expression[0][1]
        condition_translated = _translate_condition(condition)

        statement_str = f'IF ({condition_translated}) {symbol} = {value}'
        return statement_str
    elif all(len(e.args) == 0 for e in expressions) and not has_else:
        return _translate_sympy_single(symbol, expression)
    else:
        return _translate_sympy_block(symbol, expression)


def _translate_sympy_single(symbol, expression):
    statement_str = ''
    for e in expression:
        value = e[0]
        condition = e[1]

        condition_translated = _translate_condition(condition)

        statement_str += f'IF ({condition_translated}) {symbol} = {value}\n'

    return statement_str


def _translate_sympy_block(symbol, expression):
    for i, e in enumerate(expression):
        value = e[0]
        condition = e[1]

        condition_translated = _translate_condition(condition)

        if i == 0:
            statement_str = f'IF ({condition_translated}) THEN\n'
        elif condition_translated == '.true.':
            statement_str += 'ELSE\n'
        else:
            statement_str += f'ELSE IF ({condition_translated}) THEN\n'

        statement_str += f'{symbol} = {value}\n'

    statement_str += 'END IF'
    return statement_str


def _translate_condition(c):
    fprn = NMTranPrinter(settings={'source_format': 'free'})
    fortran = fprn.doprint(c).replace(' ', '')
    return fortran


def _translate_sympy_sign(s):
    args = s.expression.args

    subs_dict = dict()
    for arg in args:
        if str(arg).startswith('sign'):
            sign_arg = arg.args[0]
            subs_dict[arg] = abs(sign_arg) / sign_arg

    s.subs(subs_dict)
    fprn = NMTranPrinter(settings={'source_format': 'free'})
    fortran = fprn.doprint(s.expression)
    expr_str = f'{s.symbol} = {fortran}'
    return expr_str


def _print_custom(assignment, rvs, trans):
    expr_ordered = _order_terms(assignment, rvs, trans)
    return f'{assignment.symbol} = {expr_ordered}'


def _order_terms(assignment, rvs, trans):
    """Order terms such that random variables are placed last. Currently only supports
    additions."""
    if not isinstance(assignment.expression, sympy.Add) or rvs is None:
        return assignment.expression

    rvs_names = [rv.name for rv in rvs]

    if trans:
        trans_rvs = {v.name: k.name for k, v in trans.items() if str(k) in rvs_names}

    expr_args = assignment.expression.args
    terms_iiv_iov, terms_ruv, terms = [], [], []

    for arg in expr_args:
        arg_symbs = [s.name for s in arg.free_symbols]
        rvs_intersect = set(rvs_names).intersection(arg_symbs)

        if trans:
            trans_intersect = set(trans_rvs.keys()).intersection(arg_symbs)
            rvs_intersect.update({trans_rvs[rv] for rv in trans_intersect})

        if rvs_intersect:
            if len(rvs_intersect) == 1:
                rv_name = list(rvs_intersect)[0]
                variability_level = rvs[rv_name].level
                if variability_level == 'RUV':
                    terms_ruv.append(arg)
                    continue
            terms_iiv_iov.append(arg)
        else:
            terms.append(arg)

    if not terms_iiv_iov and not terms_ruv:
        return assignment.expression

    def arg_len(symb):
        return len([s for s in symb.args])

    terms_iiv_iov.sort(reverse=True, key=arg_len)
    terms_ruv.sort(reverse=True, key=arg_len)
    terms += terms_iiv_iov + terms_ruv

    new_order = sympy.Add(*terms, evaluate=False)

    return MyPrinter().doprint(new_order)


class ExpressionInterpreter(lark.visitors.Interpreter):
    def visit_children(self, tree):
        """Does not visit tokens"""
        return [self.visit(child) for child in tree.children if isinstance(child, lark.Tree)]

    def expression(self, node):
        t = self.visit_children(node)

        if bool(node.find('UNARY_OP')) and str(node.tokens[0]) == '-':
            unary_factor = -1
        else:
            unary_factor = 1

        if len(t) > 2:
            ops = t[1::2]
            terms = t[2::2]
            expr = unary_factor * t[0]
            for op, term in zip(ops, terms):
                if op == '+':
                    expr += term
                elif op == '-':
                    expr -= term
                elif op == '*':
                    expr *= term
                else:  # op == '/':
                    expr /= term
        else:
            expr = unary_factor * t[0]
        return expr

    def logical_expression(self, node):
        t = self.visit_children(node)
        if len(t) > 2:
            ops = t[1::2]
            terms = t[2::2]
            expr = t[0]
            for op, term in zip(ops, terms):
                expr = op(expr, term)
            return expr
        else:
            op, expr = self.visit_children(node)
            return op(expr)

    @staticmethod
    def logical_operator(node):
        name = str(node).upper()
        if name == '==' or name == '.EQ.':
            return sympy.Eq
        elif name == '/=' or name == '.NE.':
            return sympy.Ne
        elif name == '<=' or name == '.LE.':
            return sympy.Le
        elif name == '>=' or name == '.GE.':
            return sympy.Ge
        elif name == '<' or name == '.LT.':
            return sympy.Lt
        elif name == '>' or name == '.GT.':
            return sympy.Gt
        elif name == '.AND.':
            return sympy.And
        elif name == '.OR.':
            return sympy.Or
        else:  # name == '.NOT.':
            return sympy.Not

    def func(self, node):
        func, expr = self.visit_children(node)
        return func(expr)

    def func2(self, node):
        a, p = self.visit_children(node)
        return sympy.Mod(a, p)

    @staticmethod
    def intrinsic_func(node):
        smallz = 2.8e-103
        name = str(node).upper()
        if name == "EXP" or name == "DEXP":
            return sympy.exp
        if name == "PEXP":
            return lambda x: sympy.Piecewise((sympy.exp(100), x > 100), (sympy.exp(x), True))
        elif name == "LOG":
            return sympy.log
        elif name == "PLOG":
            return lambda x: sympy.Piecewise((sympy.log(smallz), x < smallz), (sympy.log(x), True))
        elif name == "LOG10":
            return lambda x: sympy.log(x, 10)
        elif name == "PLOG10":
            return lambda x: sympy.Piecewise(
                (sympy.log(smallz, 10), x < smallz), (sympy.log(x, 10), True)
            )
        elif name == "SQRT":
            return sympy.sqrt
        elif name == "PSQRT":
            return lambda x: sympy.Piecewise((0, x < 0), (sympy.sqrt(x), True))
        elif name == "SIN":
            return sympy.sin
        elif name == "COS":
            return sympy.cos
        elif name == "ABS":
            return sympy.Abs
        elif name == "TAN" or name == "PTAN":
            return sympy.tan
        elif name == "ASIN" or name == "PASIN":
            return sympy.asin
        elif name == "ACOS" or name == "PACOS":
            return sympy.acos
        elif name == "ATAN" or name == "PATAN":
            return sympy.atan
        elif name == "INT":
            return lambda x: sympy.sign(x) * sympy.floor(sympy.Abs(x))
        elif name == "GAMLN":
            return sympy.loggamma
        elif name == "PDZ":
            return lambda x: sympy.Piecewise((1 / smallz, abs(x) < smallz), (1 / x, True))
        elif name == "PZR":
            return lambda x: sympy.Piecewise((smallz, abs(x) < smallz), (x, True))
        elif name == "PNP":
            return lambda x: sympy.Piecewise((smallz, x < smallz), (x, True))
        elif name == "PHE":
            return lambda x: sympy.Piecewise((100, x > 100), (x, True))
        elif name == "PNG":
            return lambda x: sympy.Piecewise((0, x < 0), (x, True))
        else:  # name == "PHI":
            return lambda x: (1 + sympy.erf(x) / sympy.sqrt(2)) / 2

    def power(self, node):
        b, e = self.visit_children(node)
        return b ** e

    @staticmethod
    def operator(node):
        return str(node)

    @staticmethod
    def number(node):
        s = str(node)
        try:
            return sympy.Integer(s)
        except ValueError:
            s = s.replace('d', 'E')  # Fortran special format
            s = s.replace('D', 'E')
            return sympy.Float(s)

    @staticmethod
    def symbol(node):
        name = str(node).upper()
        if name.startswith('ERR('):
            name = 'EPS' + name[3:]
        symb = symbols.symbol(name)
        return symb


def lcslen(a, b):
    # generate matrix of length of longest common subsequence for sublists of both lists
    lengths = [[0] * (len(b) + 1) for _ in range(len(a) + 1)]
    for i, x in enumerate(a):
        for j, y in enumerate(b):
            if x == y:
                lengths[i + 1][j + 1] = lengths[i][j] + 1
            else:
                lengths[i + 1][j + 1] = max(lengths[i + 1][j], lengths[i][j + 1])
    return lengths


def lcsdiff(c, x, y, i, j):
    """Print the diff using LCS length matrix using backtracking"""
    if i < 0 and j < 0:
        return
    elif i < 0:
        yield from lcsdiff(c, x, y, i, j - 1)
        yield '+', y[j]
    elif j < 0:
        yield from lcsdiff(c, x, y, i - 1, j)
        yield '-', x[i]
    elif x[i] == y[j]:
        yield from lcsdiff(c, x, y, i - 1, j - 1)
        yield None, x[i]
    elif c[i][j - 1] >= c[i - 1][j]:
        yield from lcsdiff(c, x, y, i, j - 1)
        yield '+', y[j]
    else:
        yield from lcsdiff(c, x, y, i - 1, j)
        yield '-', x[i]


def diff(old, new):
    """Get diff between a and b in order for all elements

    Optimizes by first handling equal elements from the head and tail
    Each entry is a pair of operation (+, - or None) and the element
    """
    for i, (a, b) in enumerate(zip(old, new)):
        if a == b:
            yield (None, b)
        else:
            break
    else:
        if len(old) == 0 or len(new) == 0:
            i = 0
        else:
            i += 1

    rold = old[i:]
    rnew = new[i:]

    saved = []
    for a, b in zip(reversed(rold), reversed(rnew)):
        if a == b:
            saved.append((None, b))
        else:
            break

    rold = rold[: len(rold) - len(saved)]
    rnew = rnew[: len(rnew) - len(saved)]

    c = lcslen(rold, rnew)
    for op, val in lcsdiff(c, rold, rnew, len(rold) - 1, len(rnew) - 1):
        yield op, val

    for pair in reversed(saved):
        yield pair


class CodeRecord(Record):
    def __init__(self, content, parser_class):
        self.is_updated = False
        self.rvs, self.trans = None, None
        super().__init__(content, parser_class)

    @property
    def statements(self):
        statements = self._assign_statements()
        self._statements = statements
        return statements.copy()

    @statements.setter
    def statements(self, new):
        try:
            old = self._statements
        except AttributeError:
            old = self.statements
        if new == old:
            return
        old_index = 0
        node_index = 0
        kept = []
        new_nodes = []
        defined_symbols = set()  # Set of all defined symbols updated so far
        for op, s in diff(old, new):
            while (
                node_index < len(self.root.children)
                and old_index < len(self.nodes)
                and self.root.children[node_index] is not self.nodes[old_index]
            ):
                node = self.root.children[node_index]
                kept.append(node)
                node_index += 1
            if op == '+':
                statement_str = nmtran_assignment_string(s, defined_symbols, self.rvs, self.trans)
                node_tree = CodeRecordParser(statement_str).root
                for node in node_tree.all('statement'):
                    if node_index == 0:
                        node.children.insert(0, AttrToken('LF', '\n'))
                    if (
                        not node.all('LF')
                        and node_index != 0
                        or len(self.root.children) > 0
                        and self.root.children[0].rule != 'empty_line'
                    ):
                        node.children.append(AttrToken('LF', '\n'))
                    new_nodes.append(node)
                    kept.append(node)
                defined_symbols.add(s.symbol)
            elif op == '-':
                node_index += 1
                old_index += 1
            else:
                kept.append(self.root.children[node_index])
                new_nodes.append(self.root.children[node_index])
                node_index += 1
                old_index += 1
                defined_symbols.add(s.symbol)
        if node_index < len(self.root.children):  # Remaining non-statements
            kept.extend(self.root.children[node_index:])
        self.root.children = kept
        self.nodes = new_nodes
        self._statements = new.copy()

    def _assign_statements(self):
        s = []
        self.nodes = []
        for statement in self.root.all('statement'):
            for node in statement.children:
                if node.rule == 'assignment':
                    name = str(node.variable).upper()
                    expr = ExpressionInterpreter().visit(node.expression)
                    ass = Assignment(name, expr)
                    s.append(ass)
                    self.nodes.append(statement)
                elif node.rule == 'logical_if':
                    logic_expr = ExpressionInterpreter().visit(node.logical_expression)
                    try:
                        assignment = node.assignment
                    except NoSuchRuleException:
                        pass
                    else:
                        name = str(assignment.variable).upper()
                        expr = ExpressionInterpreter().visit(assignment.expression)
                        # Check if symbol was previously declared
                        else_val = sympy.Integer(0)
                        for prevass in s:
                            if prevass.symbol.name == name:
                                else_val = sympy.Symbol(name)
                                break
                        pw = sympy.Piecewise((expr, logic_expr), (else_val, True))
                        ass = Assignment(name, pw)
                        s.append(ass)
                    self.nodes.append(statement)
                elif node.rule == 'block_if':
                    interpreter = ExpressionInterpreter()
                    blocks = []  # [(logic, [(symb1, expr1), ...]), ...]
                    symbols = OrderedSet()

                    first_logic = interpreter.visit(node.block_if_start.logical_expression)
                    first_block = node.block_if_start
                    first_symb_exprs = []
                    for ifstat in first_block.all('statement'):
                        for assign_node in ifstat.all('assignment'):
                            name = str(assign_node.variable).upper()
                            first_symb_exprs.append(
                                (name, interpreter.visit(assign_node.expression))
                            )
                            symbols.add(name)
                    blocks.append((first_logic, first_symb_exprs))

                    else_if_blocks = node.all('block_if_elseif')
                    for elseif in else_if_blocks:
                        logic = interpreter.visit(elseif.logical_expression)
                        elseif_symb_exprs = []
                        for elseifstat in elseif.all('statement'):
                            for assign_node in elseifstat.all('assignment'):
                                name = str(assign_node.variable).upper()
                                elseif_symb_exprs.append(
                                    (name, interpreter.visit(assign_node.expression))
                                )
                                symbols.add(name)
                        blocks.append((logic, elseif_symb_exprs))

                    else_block = node.find('block_if_else')
                    if else_block:
                        else_symb_exprs = []
                        for elsestat in else_block.all('statement'):
                            for assign_node in elsestat.all('assignment'):
                                name = str(assign_node.variable).upper()
                                else_symb_exprs.append(
                                    (name, interpreter.visit(assign_node.expression))
                                )
                                symbols.add(name)
                        piecewise_logic = True
                        if len(blocks[0][1]) == 0 and not else_if_blocks:
                            # Special case for empty if
                            piecewise_logic = sympy.Not(blocks[0][0])
                        blocks.append((piecewise_logic, else_symb_exprs))

                    for symbol in symbols:
                        pairs = []
                        for block in blocks:
                            logic = block[0]
                            for cursymb, expr in block[1]:
                                if cursymb == symbol:
                                    pairs.append((expr, logic))
                        pw = sympy.Piecewise(*pairs)
                        ass = Assignment(symbol, pw)
                        s.append(ass)
                    self.nodes.append(statement)

        statements = ModelStatements(s)
        return statements

    def from_odes(self, ode_system):
        """Set statements of record given an eplicit ode system"""
        odes = ode_system.odes[:-1]  # Skip last ode as it is for the output compartment
        functions = [ode.lhs.args[0] for ode in odes]
        function_map = {f: symbols.symbol(f'A({i + 1})') for i, f in enumerate(functions)}
        statements = []
        for i, ode in enumerate(odes):
            # For now Piecewise signals zero-order infusions, which are handled with parameters
            ode = ode.replace(sympy.Piecewise, lambda a1, a2: 0)
            symbol = symbols.symbol(f'DADT({i + 1})')
            expression = ode.rhs.subs(function_map)
            statements.append(Assignment(symbol, expression))
        self.statements = statements

    def __str__(self):
        if self.is_updated:
            s = str(self.root)
            newlines = []
            # FIXME: Workaround for upper casing all code but not comments.
            # should properly be handled in a custom printer
            for line in s.split('\n'):
                parts = line.split(';', 1)
                modline = parts[0].upper()
                if len(parts) == 2:
                    modline += ';' + parts[1]
                newlines.append(modline)
            return self.raw_name + '\n'.join(newlines)
        return super(CodeRecord, self).__str__()



import pytest

import pharmpy.config as config
import pharmpy.data as data
from pharmpy.config import ConfigurationContext
from pharmpy.plugins.nonmem import conf


def test_config_item():
    class TestClass:
        item1 = config.ConfigItem(23, "A single number")

    obj = TestClass()
    assert obj.item1 == 23
    obj.item1 = 91
    assert obj.item1 == 91
    with pytest.raises(TypeError):
        obj.item1 = "A"


def test_data_config():
    assert data.conf.na_values == [-99]


def test_read_config():
    config.read_configuration()
    assert len(conf.parameter_names) <= 3


def test_str(pheno_path, datadir, fs):
    with ConfigurationContext(conf, parameter_names=['abbr', 'basic']):
        assert str(conf) == "parameter_names:\t['abbr', 'basic']\n"



# The NONMEM Model clas

import copy
import re
import shutil
import warnings
from io import StringIO
from pathlib import Path

import pharmpy.data
import pharmpy.model
import pharmpy.model_factory
import pharmpy.plugins.nonmem
from pharmpy.data import DatasetError
from pharmpy.estimation import EstimationMethod, list_supported_est
from pharmpy.model import ModelSyntaxError
from pharmpy.parameter import Parameters
from pharmpy.plugins.nonmem.results import NONMEMChainedModelfitResults
from pharmpy.plugins.nonmem.table import NONMEMTableFile, PhiTable
from pharmpy.random_variables import RandomVariables
from pharmpy.statements import Assignment, CompartmentalSystem, ODESystem
from pharmpy.symbols import symbol as S
from pharmpy.workflows import NullModelDatabase, default_model_database

from .advan import compartmental_model
from .nmtran_parser import NMTranParser
from .records.factory import create_record
from .update import (
    update_abbr_record,
    update_ccontra,
    update_estimation,
    update_parameters,
    update_random_variables,
    update_statements,
)


def detect_model(src, *args, **kwargs):
    """Check if src represents a NONMEM control stream
    i.e. check if it is a file that contain $PRO
    """
    if not isinstance(src, str):
        return None
    is_control_stream = re.search(r'^\s*\$PRO', src, re.MULTILINE)
    if is_control_stream:
        return Model
    else:
        return None


def convert_model(model):
    """Convert any model into a NONMEM model"""
    if isinstance(model, Model):
        return model
    model = model.to_generic_model()
    code = '$PROBLEM\n'
    code += '$INPUT ' + ' '.join(model.dataset.columns) + '\n'
    code += '$DATA file.csv IGNORE=@\n'
    if model.statements.ode_system is None:
        code += '$PRED\n'
    else:
        code += '$PK\n'
    nm_model = pharmpy.model_factory.Model(StringIO(code))
    nm_model.random_variables = model.random_variables
    nm_model.parameters = model.parameters
    nm_model.statements = model.statements
    if hasattr(model, 'name'):
        nm_model.name = model.name
    if hasattr(model, 'dependent_variable'):
        nm_model.dependent_variable = model.dependent_variable
    nm_model._data_frame = model.dataset
    nm_model._estimation_steps = model.estimation_steps
    nm_model.update_source()
    try:
        nm_model.database = model.database
    except AttributeError:
        pass
    return nm_model


class Model(pharmpy.model.Model):
    def __init__(self, code, path=None, **kwargs):
        super().__init__()
        parser = NMTranParser()
        if path is None:
            self._name = 'run1'
            self.database = NullModelDatabase()
            self.filename_extension = '.ctl'
        else:
            self._name = path.stem
            self.database = default_model_database(path=path.parent)
            self.filename_extension = path.suffix
        self.control_stream = parser.parse(code)
        self._initial_individual_estimates_updated = False
        self._updated_etas_file = None
        self._dataset_updated = False
        self._modelfit_results = None
        self.dependent_variable = S('Y')
        self.observation_transformation = self.dependent_variable
        self._old_observation_transformation = self.dependent_variable

    def __eq__(self, other):
        # Compare NONMEM code to see if models are the same
        self.update_source()
        other.update_source()
        return str(self) == str(other)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        m = re.search(r'.*?(\d+)$', new_name)
        if m:
            n = int(m.group(1))
            for table in self.control_stream.get_records('TABLE'):
                table_path = table.path
                table_name = table_path.stem
                m = re.search(r'(.*?)(\d+)$', table_name)
                if m:
                    table_stem = m.group(1)
                    new_table_name = f'{table_stem}{n}'
                    table.path = table_path.parent / new_table_name
        self._name = new_name

    @property
    def modelfit_results(self):
        if self._modelfit_results is None:
            self.read_modelfit_results()
        return self._modelfit_results

    @modelfit_results.setter
    def modelfit_results(self, res):
        self._modelfit_results = res

    def update_source(self, path=None, force=False, nofiles=False):
        """Update the source

        path - path to modelfile
        nofiles - Set to not write any files (i.e. dataset, phi input etc)
        """
        self._update_initial_individual_estimates(path, nofiles)
        if hasattr(self, '_random_variables'):
            update_random_variables(self, self._old_random_variables, self._random_variables)
            self._old_random_variables = self._random_variables.copy()
        if hasattr(self, '_parameters'):
            update_parameters(self, self._old_parameters, self._parameters)
            self._old_parameters = self._parameters.copy()
        trans = self.parameter_translation(reverse=True, remove_idempotent=True, as_symbols=True)
        rv_trans = self.rv_translation(reverse=True, remove_idempotent=True, as_symbols=True)
        trans.update(rv_trans)
        if pharmpy.plugins.nonmem.conf.write_etas_in_abbr:
            abbr_trans = self._abbr_translation(rv_trans)
            trans.update(abbr_trans)
        if trans:
            self.statements  # Read statements unless read
        if hasattr(self, '_statements'):
            update_statements(self, self._old_statements, self._statements, trans)
            self._old_statements = self._statements.copy()

        if self._dataset_updated:
            # FIXME: If no name set use the model name. Set that when setting dataset to input!
            if not nofiles:
                if path is not None:
                    dir_path = path.parent
                else:
                    dir_path = None
                datapath = self.dataset.pharmpy.write_csv(path=dir_path, force=force)
                self.dataset_path = datapath.name

            data_record = self.control_stream.get_records('DATA')[0]

            label = self.dataset.columns[0]
            data_record.ignore_character_from_header(label)
            self._update_input(self.dataset.columns)

            # Remove IGNORE/ACCEPT. Could do diff between old dataset and find simple
            # IGNOREs to add i.e. for filter out certain ID.
            del data_record.ignore
            del data_record.accept
            self._dataset_updated = False

        self._update_sizes()
        update_estimation(self)

        if self.observation_transformation != self._old_observation_transformation:
            if not nofiles:
                update_ccontra(self, path, force)

    def _abbr_translation(self, rv_trans):
        abbr_pharmpy = self.control_stream.abbreviated.translate_to_pharmpy_names()
        abbr_replace = self.control_stream.abbreviated.replace
        abbr_trans = update_abbr_record(self, rv_trans)
        abbr_recs = {
            S(abbr_pharmpy[value]): S(key)
            for key, value in abbr_replace.items()
            if value in abbr_pharmpy.keys()
        }
        abbr_trans.update(abbr_recs)
        return abbr_trans

    def _update_sizes(self):
        """Update $SIZES if needed"""
        all_sizes = self.control_stream.get_records('SIZES')
        if len(all_sizes) == 0:
            sizes = create_record('$SIZES ')
        else:
            sizes = all_sizes[0]
        odes = self.statements.ode_system
        if odes is not None and isinstance(odes, CompartmentalSystem):
            n_compartments = len(odes)
            sizes.PC = n_compartments
        if len(all_sizes) == 0 and len(str(sizes)) > 7:
            self.control_stream.insert_record(str(sizes))

    def _update_initial_individual_estimates(self, path, nofiles=False):
        """Update $ETAS

        Could have 0 FIX in model. Need to read these
        """
        if path is None:  # What to do here?
            phi_path = Path('.')
        else:
            phi_path = path.parent
        phi_path /= f'{self.name}_input.phi'

        if self._initial_individual_estimates_updated:
            etas = self.initial_individual_estimates
            zero_fix = self._zero_fix_rvs(eta=True)
            if zero_fix:
                for eta in zero_fix:
                    etas[eta] = 0
            etas = self._sort_eta_columns(etas)
            if not nofiles:
                phi = PhiTable(df=etas)
                table_file = NONMEMTableFile(tables=[phi])
                table_file.write(phi_path)
            # FIXME: This is a common operation
            eta_records = self.control_stream.get_records('ETAS')
            if eta_records:
                record = eta_records[0]
            else:
                record = self.control_stream.append_record('$ETAS ')
            record.path = phi_path
        elif self._updated_etas_file:
            eta_records = self.control_stream.get_records('ETAS')
            if eta_records:
                record = eta_records[0]
            else:
                record = self.control_stream.append_record('$ETAS')
            shutil.copy(self._updated_etas_file, phi_path)
            record.path = phi_path

        if self._initial_individual_estimates_updated or self._updated_etas_file:
            first_est_record = self.control_stream.get_records('ESTIMATION')[0]
            try:
                first_est_record.option_pairs['MCETA']
            except KeyError:
                first_est_record.set_option('MCETA', 1)
            self._updated_etas_file = None
            self._initial_individual_estimates_updated = False

    def validate(self):
        """Validates NONMEM model (records) syntactically."""
        self.control_stream.validate()

    @property
    def parameters(self):
        """Get the Parameters of all parameters"""
        try:
            return self._parameters
        except AttributeError:
            pass

        self._read_parameters()
        if (
            'comment' in pharmpy.plugins.nonmem.conf.parameter_names
            or 'abbr' in pharmpy.plugins.nonmem.conf.parameter_names
        ):
            self.statements
            # reading statements might change parameters. Resetting _old_parameters
            self._old_parameters = self._parameters.copy()
        return self._parameters

    def _read_parameters(self):
        next_theta = 1
        params = Parameters()
        for theta_record in self.control_stream.get_records('THETA'):
            thetas = theta_record.parameters(next_theta, seen_labels=set(params.names))
            params.extend(thetas)
            next_theta += len(thetas)
        next_omega = 1
        previous_size = None
        for omega_record in self.control_stream.get_records('OMEGA'):
            omegas, next_omega, previous_size = omega_record.parameters(
                next_omega, previous_size, seen_labels=set(params.names)
            )
            params.extend(omegas)
        next_sigma = 1
        previous_size = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            sigmas, next_sigma, previous_size = sigma_record.parameters(
                next_sigma, previous_size, seen_labels=set(params.names)
            )
            params.extend(sigmas)
        self._parameters = params
        self._old_parameters = params.copy()

    @parameters.setter
    def parameters(self, params):
        """params can be a Parameters or a dict-like with name: value

        Current restrictions:
         * Parameters only supported for new initial estimates and fix
         * Only set the exact same parameters. No additions and no removing of parameters
        """
        if isinstance(params, Parameters):
            inits = params.inits
        else:
            inits = params
        if not self.random_variables.validate_parameters(inits):
            raise ValueError("New parameter inits are not valid")

        self.parameters
        if not isinstance(params, Parameters):
            self._parameters.inits = params
        else:
            self._parameters = params

    @property
    def initial_individual_estimates(self):
        """Initial individual estimates

        These are taken from the $ETAS FILE. 0 FIX ETAs are removed.
        If no $ETAS is present None will be returned.

        Setter assumes that all IDs are present
        """
        try:
            return self._initial_individual_estimates
        except AttributeError:
            pass
        etas = self.control_stream.get_records('ETAS')
        if etas:
            path = Path(etas[0].path)
            if not path.is_absolute():
                source_dir = self.database.retrieve_file(
                    self.name, self.name + self.filename_extension
                ).parent
                path = source_dir / path
                path = path.resolve()
            phi_tables = NONMEMTableFile(path)
            rv_names = [rv.name for rv in self.random_variables if rv.name.startswith('ETA')]
            etas = next(phi_tables).etas[rv_names]
            self._initial_individual_estimates = etas
        else:
            self._initial_individual_estimates = None
        return self._initial_individual_estimates

    @initial_individual_estimates.setter
    def initial_individual_estimates(self, estimates):
        rv_names = {rv.name for rv in self.random_variables if rv.name.startswith('ETA')}
        columns = set(estimates.columns)
        if columns < rv_names:
            raise ValueError(
                f'Cannot set initial estimate for random variable not in the model:'
                f' {rv_names - columns}'
            )
        diff = columns - rv_names
        # If not setting all etas automatically set remaining to 0 for all individuals
        if len(diff) > 0:
            for name in diff:
                estimates = estimates.copy(deep=True)
                estimates[name] = 0
            estimates = self._sort_eta_columns(estimates)
        self._initial_individual_estimates = estimates
        self._initial_individual_estimates_updated = True
        self._updated_etas_file = None

    def update_individual_estimates(self, source):
        """Update initial individual estimates from another model"""
        self._initial_individual_estimates_updated = False
        try:
            del self._initial_individual_estimates
        except AttributeError:
            pass
        path = self.database.retrieve_file(source.name, source.name + source.filename_extension)
        self._updated_etas_file = path.parent / (source.name + '.phi')

    def _sort_eta_columns(self, df):
        return df.reindex(sorted(df.columns), axis=1)

    @property
    def statements(self):
        try:
            return self._statements
        except AttributeError:
            pass

        rec = self.get_pred_pk_record()
        statements = rec.statements

        des = self._get_des_record()
        error = self._get_error_record()
        if error:
            sub = self.control_stream.get_records('SUBROUTINES')[0]
            advan = sub.get_option_startswith('ADVAN')
            trans = sub.get_option_startswith('TRANS')
            if not trans:
                trans = 'TRANS1'
            comp = compartmental_model(self, advan, trans, des)
            if comp is not None:
                cm, link = comp
                statements += [cm, link]
            else:
                statements.append(ODESystem())  # FIXME: Placeholder for ODE-system
                # FIXME: Dummy link statement
                statements.append(Assignment('F', S('F')))
            statements += error.statements

        if not hasattr(self, '_parameters'):
            self._read_parameters()

        trans_statements, trans_params = self._create_name_trans(statements)
        for key, value in trans_params.items():
            try:
                self.parameters[key].name = value
                for theta in self.control_stream.get_records('THETA'):
                    theta.update_name_map(trans_params)
                for omega in self.control_stream.get_records('OMEGA'):
                    omega.update_name_map(trans_params)
                for sigma in self.control_stream.get_records('SIGMA'):
                    sigma.update_name_map(trans_params)
            except KeyError:
                self.random_variables.subs({S(key): value})

        statements.subs(trans_statements)

        self._statements = statements
        self._old_statements = statements.copy()
        return statements

    @statements.setter
    def statements(self, statements_new):
        self.statements  # Read in old statements
        self._statements = statements_new

    def get_pred_pk_record(self):
        pred = self.control_stream.get_records('PRED')

        if not pred:
            pk = self.control_stream.get_records('PK')
            if not pk:
                raise ModelSyntaxError('Model has no $PK or $PRED')
            return pk[0]
        else:
            return pred[0]

    def _get_pk_record(self):
        pk = self.control_stream.get_records('PK')
        if pk:
            pk = pk[0]
        return pk

    def _get_error_record(self):
        error = self.control_stream.get_records('ERROR')
        if error:
            error = error[0]
        return error

    def _get_des_record(self):
        des = self.control_stream.get_records('DES')
        if des:
            des = des[0]
        return des

    def _create_name_trans(self, statements):
        rvs = self.random_variables

        conf_functions = {
            'comment': self._name_as_comments(statements),
            'abbr': self._name_as_abbr(rvs),
            'basic': self._name_as_basic(),
        }

        abbr = self.control_stream.abbreviated.replace
        pset_current = {
            **self.parameter_translation(reverse=True),
            **{rv: rv for rv in [rv.name for rv in rvs]},
        }
        sset_current = {
            **abbr,
            **{
                rv.name: rv.name
                for rv in rvs
                if rv.name not in abbr.keys() and rv.symbol in statements.free_symbols
            },
            **{
                p: p
                for p in pset_current.values()
                if p not in abbr.keys() and S(p) in statements.free_symbols
            },
        }

        trans_sset, trans_pset = dict(), dict()
        names_sset_translated, names_pset_translated, names_basic = [], [], []
        clashing_symbols = set()

        for setting in pharmpy.plugins.nonmem.conf.parameter_names:
            trans_sset_setting, trans_pset_setting = conf_functions[setting]
            if setting != 'basic':
                clashing_symbols.update(
                    self._clashing_symbols(statements, {**trans_sset_setting, **trans_pset_setting})
                )
            for name_current, name_new in trans_sset_setting.items():
                name_nonmem = sset_current[name_current]

                if S(name_new) in clashing_symbols or name_nonmem in names_sset_translated:
                    continue

                name_in_sset_current = {v: k for k, v in sset_current.items()}[name_nonmem]
                trans_sset[name_in_sset_current] = name_new
                names_sset_translated.append(name_nonmem)

                if name_nonmem in pset_current.values() and name_new in pset_current.keys():
                    names_pset_translated.append(name_nonmem)

            for name_current, name_new in trans_pset_setting.items():
                name_nonmem = pset_current[name_current]

                if S(name_new) in clashing_symbols or name_nonmem in names_pset_translated:
                    continue

                trans_pset[name_current] = name_new
                names_pset_translated.append(name_nonmem)

            if setting == 'basic':
                params_left = [k for k in pset_current.keys() if k not in names_pset_translated]
                params_left += [rv.name for rv in rvs if rv.name not in names_sset_translated]
                names_basic = [name for name in params_left if name not in names_sset_translated]
                break

        if clashing_symbols:
            warnings.warn(
                f'The parameter names {clashing_symbols} are also names of variables '
                f'in the model code. Falling back to the in naming scheme config '
                f'names for these.'
            )

        names_nonmem_all = [rv.name for rv in rvs] + [
            key for key in self.parameter_translation().keys()
        ]

        if set(names_nonmem_all) - set(names_sset_translated + names_pset_translated + names_basic):
            raise ValueError(
                'Mismatch in number of parameter names, all have not been accounted for. If basic '
                'NONMEM-names are desired as fallback, double-check that "basic" is included in '
                'config-settings for parameter_names.'
            )
        return trans_sset, trans_pset

    def _name_as_comments(self, statements):
        params_current = self.parameter_translation(remove_idempotent=True)
        for name_abbr, name_nonmem in self.control_stream.abbreviated.replace.items():
            if name_nonmem in params_current.keys():
                params_current[name_abbr] = params_current.pop(name_nonmem)
        trans_params = {
            name_comment: name_comment
            for name_current, name_comment in params_current.items()
            if S(name_current) not in statements.free_symbols
        }
        trans_statements = {
            name_current: name_comment
            for name_current, name_comment in params_current.items()
            if S(name_current) in statements.free_symbols
        }
        return trans_statements, trans_params

    def _name_as_abbr(self, rvs):
        pharmpy_names = self.control_stream.abbreviated.translate_to_pharmpy_names()
        params_current = self.parameter_translation(remove_idempotent=True, reverse=True)
        trans_params = {
            name_nonmem: name_abbr
            for name_nonmem, name_abbr in pharmpy_names.items()
            if name_nonmem in self.parameter_translation().keys()
            or name_nonmem in [rv.name for rv in rvs]
        }
        for name_nonmem, name_abbr in params_current.items():
            if name_abbr in trans_params.keys():
                trans_params[name_nonmem] = trans_params.pop(name_abbr)
        trans_statements = {
            name_abbr: pharmpy_names[name_nonmem]
            for name_abbr, name_nonmem in self.control_stream.abbreviated.replace.items()
        }
        return trans_statements, trans_params

    def _name_as_basic(self):
        trans_params = {
            name_current: name_nonmem
            for name_current, name_nonmem in self.parameter_translation(reverse=True).items()
            if name_current != name_nonmem
        }
        trans_statements = self.control_stream.abbreviated.replace
        return trans_statements, trans_params

    @staticmethod
    def _clashing_symbols(statements, trans_statements):
        parameter_symbols = {S(symb) for _, symb in trans_statements.items()}
        clashing_symbols = parameter_symbols & statements.free_symbols
        return clashing_symbols

    def replace_abbr(self, replace):
        for key, value in replace.items():
            try:
                self.parameters[key].name = value
            except KeyError:
                pass
        self.random_variables.rename(replace)

    def _zero_fix_rvs(self, eta=True):
        zero_fix = []
        if eta:
            prev_cov = None
            next_omega = 1
            for omega_record in self.control_stream.get_records('OMEGA'):
                _, next_omega, prev_cov, new_zero_fix = omega_record.random_variables(
                    next_omega, prev_cov
                )
                zero_fix += new_zero_fix
        else:
            prev_cov = None
            next_sigma = 1
            for sigma_record in self.control_stream.get_records('SIGMA'):
                _, next_sigma, prev_cov, new_zero_fix = sigma_record.random_variables(
                    next_sigma, prev_cov
                )
                zero_fix += new_zero_fix
        return zero_fix

    @property
    def random_variables(self):
        try:
            return self._random_variables
        except AttributeError:
            pass
        rvs = RandomVariables()
        next_omega = 1
        prev_cov = None

        for omega_record in self.control_stream.get_records('OMEGA'):
            etas, next_omega, prev_cov, _ = omega_record.random_variables(next_omega, prev_cov)
            rvs.extend(etas)
        self.adjust_iovs(rvs)
        next_sigma = 1
        prev_cov = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            epsilons, next_sigma, prev_cov, _ = sigma_record.random_variables(next_sigma, prev_cov)
            rvs.extend(epsilons)
        self._random_variables = rvs
        self._old_random_variables = rvs.copy()

        if (
            'comment' in pharmpy.plugins.nonmem.conf.parameter_names
            or 'abbr' in pharmpy.plugins.nonmem.conf.parameter_names
        ):
            self.statements

        return self._random_variables

    @staticmethod
    def adjust_iovs(rvs):
        for i, rv in enumerate(rvs):
            try:
                next_rv = rvs[i + 1]
            except IndexError:
                break

            if rv.level != 'IOV' and next_rv.level == 'IOV':
                associated_rvs = rv.joint_names
                if len(associated_rvs) > 1:
                    for rv_con in associated_rvs:
                        rvs[rv_con].level = 'IOV'
                else:
                    rv.level = 'IOV'

    @random_variables.setter
    def random_variables(self, new):
        self.random_variables  # Read in old random variables
        self._random_variables = new.copy()

    @property
    def estimation_steps(self):
        try:
            return self._estimation_steps
        except AttributeError:
            pass

        steps = []
        records = self.control_stream.get_records('ESTIMATION')
        covrec = self.control_stream.get_records('COVARIANCE')
        for record in records:
            value = record.get_option('METHOD')
            if value is None or value == '0' or value == 'ZERO':
                name = 'fo'
            elif value == '1' or value == 'CONDITIONAL' or value == 'COND':
                name = 'foce'
            else:
                if value in list_supported_est():
                    name = value
                else:
                    raise ModelSyntaxError(
                        f'Non-recognized estimation method in: {str(record.root)}'
                    )

            interaction = False
            evaluation = False
            maximum_evaluations = None
            cov = False
            laplace = False

            if record.has_option('INTERACTION') or record.has_option('INTER'):
                interaction = True
            maxeval_opt = (
                record.get_option('MAXEVAL') if not None else record.get_option('MAXEVALS')
            )
            if maxeval_opt is not None:
                if (name.upper() == 'FO' or name.upper() == 'FOCE') and int(maxeval_opt) == 0:
                    evaluation = True
                else:
                    maximum_evaluations = int(maxeval_opt)
            eval_opt = record.get_option('EONLY')
            if eval_opt is not None and int(eval_opt) == 1:
                evaluation = True
            if covrec:
                cov = True
            if record.has_option('LAPLACIAN') or record.has_option('LAPLACE'):
                laplace = True
            protected_names = [
                name.upper(),
                'EONLY',
                'INTERACTION',
                'INTER',
                'LAPLACE',
                'LAPLACIAN',
                'MAXEVAL',
                'MAXEVALS',
                'METHOD',
                'METH',
            ]

            tool_options = {
                option.key: option.value
                for option in record.all_options
                if option.key not in protected_names
            }
            if not tool_options:
                tool_options = None

            meth = EstimationMethod(
                name,
                interaction=interaction,
                cov=cov,
                evaluation=evaluation,
                maximum_evaluations=maximum_evaluations,
                laplace=laplace,
                tool_options=tool_options,
            )
            steps.append(meth)

        self._estimation_steps = steps
        self._old_estimation_steps = copy.deepcopy(steps)
        return steps

    @property
    def model_code(self):
        self.update_source(nofiles=True)
        return str(self.control_stream)

    @property
    def dataset_path(self):
        """This property is NONMEM specific"""
        record = self.control_stream.get_records('DATA')[0]
        path = Path(record.filename)
        if not path.is_absolute():
            path = self.database.retrieve_file(self.name, path)
        try:
            return path.resolve()
        except FileNotFoundError:
            return path

    @dataset_path.setter
    def dataset_path(self, path):
        path = Path(path)
        assert (
            not path.exists() or path.is_file()
        ), 'input path change, but non-file exists at ' 'target (%s)' % str(path)
        record = self.control_stream.get_records('DATA')[0]
        record.filename = str(path)

    @property
    def dataset(self):
        try:
            return self._data_frame
        except AttributeError:
            self._data_frame = self._read_dataset(raw=False)
        return self._data_frame

    @dataset.setter
    def dataset(self, df):
        self._dataset_updated = True
        self._data_frame = df

    def read_raw_dataset(self, parse_columns=tuple()):
        return self._read_dataset(raw=True, parse_columns=parse_columns)

    @staticmethod
    def _synonym(key, value):
        """Return a tuple reserved name and synonym"""
        _reserved_column_names = [
            'ID',
            'L1',
            'L2',
            'DV',
            'MDV',
            'RAW_',
            'MRG_',
            'RPT_',
            'TIME',
            'DATE',
            'DAT1',
            'DAT2',
            'DAT3',
            'EVID',
            'AMT',
            'RATE',
            'SS',
            'II',
            'ADDL',
            'CMT',
            'PCMT',
            'CALL',
            'CONT',
        ]
        if key in _reserved_column_names:
            return (key, value)
        elif value in _reserved_column_names:
            return (value, key)
        else:
            raise DatasetError(
                f'A column name "{key}" in $INPUT has a synonym to a non-reserved '
                f'column name "{value}"'
            )

    def _column_info(self):
        """List all column names in order.
        Use the synonym when synonym exists.
        return tuple of three lists, colnames, coltypes and drop together with a dictionary
        of replacements for reserved names (aka synonyms).
        Anonymous columns, i.e. DROP or SKIP alone, will be given unique names _DROP1, ...
        """
        input_records = self.control_stream.get_records("INPUT")
        colnames = []
        coltypes = []
        drop = []
        synonym_replacement = {}
        next_anonymous = 1
        for record in input_records:
            for key, value in record.all_options:
                if value:
                    if key == 'DROP' or key == 'SKIP':
                        drop.append(True)
                        colnames.append(value)
                        reserved_name = value
                    elif value == 'DROP' or value == 'SKIP':
                        colnames.append(key)
                        drop.append(True)
                        reserved_name = key
                    else:
                        drop.append(False)
                        (reserved_name, synonym) = Model._synonym(key, value)
                        synonym_replacement[reserved_name] = synonym
                        colnames.append(synonym)
                else:
                    if key == 'DROP' or key == 'SKIP':
                        drop.append(True)
                        name = f'_DROP{next_anonymous}'
                        colnames.append(name)
                        reserved_name = name
                        next_anonymous += 1
                    else:
                        drop.append(False)
                        colnames.append(key)
                        reserved_name = key
                coltypes.append(pharmpy.data.read.infer_column_type(reserved_name))
        return colnames, coltypes, drop, synonym_replacement

    def _update_input(self, new_names):
        """Update $INPUT with new column names

        currently supporting append columns at end and removing columns
        """
        colnames, _, _, _ = self._column_info()
        removed_columns = set(colnames) - set(new_names)
        input_records = self.control_stream.get_records("INPUT")
        for col in removed_columns:
            input_records[0].remove_option(col)
        appended_names = new_names[len(colnames) :]
        last_input_record = input_records[-1]
        for colname in appended_names:
            last_input_record.append_option(colname)

    def _replace_synonym_in_filters(filters, replacements):
        result = []
        for f in filters:
            if f.COLUMN in replacements:
                s = ''
                for child in f.children:
                    if child.rule == 'COLUMN':
                        value = replacements[f.COLUMN]
                    else:
                        value = str(child)
                    s += value
            else:
                s = str(f)
            result.append(s)
        return result

    def _read_dataset(self, raw=False, parse_columns=tuple()):
        data_records = self.control_stream.get_records('DATA')
        ignore_character = data_records[0].ignore_character
        null_value = data_records[0].null_value
        (colnames, coltypes, drop, replacements) = self._column_info()

        if raw:
            ignore = None
            accept = None
        else:
            # FIXME: All direct handling of control stream spanning
            # over one or more records should move
            ignore = data_records[0].ignore
            accept = data_records[0].accept
            # FIXME: This should really only be done if setting the dataset
            if ignore:
                ignore = Model._replace_synonym_in_filters(ignore, replacements)
            else:
                accept = Model._replace_synonym_in_filters(accept, replacements)

        df = pharmpy.data.read_nonmem_dataset(
            self.dataset_path,
            raw,
            ignore_character,
            colnames,
            coltypes,
            drop,
            null_value=null_value,
            parse_columns=parse_columns,
            ignore=ignore,
            accept=accept,
        )
        if self._get_pk_record():
            if 'EVID' in df.columns:
                if 'MDV' in df.columns:
                    df.pharmpy.column_type['MDV'] = pharmpy.data.ColumnType.UNKNOWN
                df.pharmpy.column_type['EVID'] = pharmpy.data.ColumnType.EVENT
            if 'AMT' in df.columns:
                df.pharmpy.column_type['AMT'] = pharmpy.data.ColumnType.DOSE
        # Let TIME be the idv in both $PK and $PRED models
        idvcol = replacements.get('TIME', 'TIME')
        if idvcol in df.columns:
            df.pharmpy.column_type[idvcol] = pharmpy.data.ColumnType.IDV
        df.name = self.dataset_path.stem

        # Remove individuals without observations
        try:
            have_obs = set(df.pharmpy.observations.index.unique(level=0))
        except DatasetError:
            pass
        else:
            all_ids = set(df.pharmpy.ids)
            ids_to_remove = all_ids - have_obs
            df = df[~df['ID'].isin(ids_to_remove)]
        return df

    def rv_translation(self, reverse=False, remove_idempotent=False, as_symbols=False):
        self.random_variables
        d = dict()
        for record in self.control_stream.get_records('OMEGA'):
            for key, value in record.eta_map.items():
                nonmem_name = f'ETA({value})'
                d[nonmem_name] = key
        for record in self.control_stream.get_records('SIGMA'):
            for key, value in record.eta_map.items():
                nonmem_name = f'EPS({value})'
                d[nonmem_name] = key
        if remove_idempotent:
            d = {key: val for key, val in d.items() if key != val}
        if reverse:
            d = {val: key for key, val in d.items()}
        if as_symbols:
            d = {S(key): S(val) for key, val in d.items()}
        return d

    def parameter_translation(self, reverse=False, remove_idempotent=False, as_symbols=False):
        """Get a dict of NONMEM name to Pharmpy parameter name
        i.e. {'THETA(1)': 'TVCL', 'OMEGA(1,1)': 'IVCL'}
        """
        self.parameters
        d = dict()
        for theta_record in self.control_stream.get_records('THETA'):
            for key, value in theta_record.name_map.items():
                nonmem_name = f'THETA({value})'
                d[nonmem_name] = key
        for record in self.control_stream.get_records('OMEGA'):
            for key, value in record.name_map.items():
                nonmem_name = f'OMEGA({value[0]},{value[1]})'
                d[nonmem_name] = key
        for record in self.control_stream.get_records('SIGMA'):
            for key, value in record.name_map.items():
                nonmem_name = f'SIGMA({value[0]},{value[1]})'
                d[nonmem_name] = key
        if remove_idempotent:
            d = {key: val for key, val in d.items() if key != val}
        if reverse:
            d = {val: key for key, val in d.items()}
        if as_symbols:
            d = {S(key): S(val) for key, val in d.items()}
        return d

    def read_modelfit_results(self):
        try:
            ext_path = self.database.retrieve_file(self.name, self.name + '.ext')
        except FileNotFoundError:
            self._modelfit_results = None
            return None
        if ext_path is not None:
            self._modelfit_results = NONMEMChainedModelfitResults(ext_path, model=self)
            return self._modelfit_results



from io import StringIO

import pytest
import sympy
from pyfakefs.fake_filesystem_unittest import Patcher

from pharmpy import Model
from pharmpy.config import ConfigurationContext
from pharmpy.estimation import EstimationMethod
from pharmpy.model import ModelSyntaxError
from pharmpy.modeling import add_iiv, set_zero_order_absorption, set_zero_order_elimination
from pharmpy.parameter import Parameter
from pharmpy.plugins.nonmem import conf, convert_model
from pharmpy.plugins.nonmem.nmtran_parser import NMTranParser
from pharmpy.random_variables import RandomVariable
from pharmpy.statements import Assignment, ModelStatements, ODESystem
from pharmpy.symbols import symbol


def S(x):
    return symbol(x)


def test_source(pheno_path):
    model = Model(pheno_path)
    assert model.model_code.startswith('$PROBLEM PHENOBARB')


def test_update_inits(pheno_path):
    model = Model(pheno_path)
    model.update_inits()

    with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
        model = Model(pheno_path)
        model.update_inits()
        model.update_source()


def test_empty_ext_file(testdata):
    # assert existing but empty ext-file does not give modelfit_results
    model = Model(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'noESTwithSIM' / 'onlysim.mod'
    )
    with pytest.raises(FileNotFoundError):
        model.database.retrieve_file(model.name, model.name + '.ext')
    assert model.modelfit_results is None


def test_detection():
    Model(StringIO("$PROBLEM this"))
    Model(StringIO("   \t$PROBLEM skld fjl"))
    Model(StringIO(" $PRO l907"))


def test_validate(pheno_path):
    model = Model(pheno_path)
    model.validate()

    model = Model(StringIO("$PROBLEM this\n$SIZES LIM1=3000"))
    with pytest.raises(ModelSyntaxError):
        model.validate()


def test_parameters(pheno_path):
    model = Model(pheno_path)
    params = model.parameters
    assert len(params) == 6
    assert model.parameters['THETA(1)'] == Parameter('THETA(1)', 0.00469307, lower=0, upper=1000000)
    assert model.parameters['THETA(2)'] == Parameter('THETA(2)', 1.00916, lower=0, upper=1000000)
    assert model.parameters['THETA(3)'] == Parameter('THETA(3)', 0.1, lower=-0.99, upper=1000000)
    assert model.parameters['OMEGA(1,1)'] == Parameter(
        'OMEGA(1,1)', 0.0309626, lower=0, upper=sympy.oo
    )
    assert model.parameters['OMEGA(2,2)'] == Parameter(
        'OMEGA(2,2)', 0.031128, lower=0, upper=sympy.oo
    )
    assert model.parameters['SIGMA(1,1)'] == Parameter(
        'SIGMA(1,1)', 0.013241, lower=0, upper=sympy.oo
    )


def test_set_parameters(pheno_path):
    model = Model(pheno_path)
    params = {
        'THETA(1)': 0.75,
        'THETA(2)': 0.5,
        'THETA(3)': 0.25,
        'OMEGA(1,1)': 0.1,
        'OMEGA(2,2)': 0.2,
        'SIGMA(1,1)': 0.3,
    }
    model.parameters = params
    assert model.parameters['THETA(1)'] == Parameter('THETA(1)', 0.75, lower=0, upper=1000000)
    assert model.parameters['THETA(2)'] == Parameter('THETA(2)', 0.5, lower=0, upper=1000000)
    assert model.parameters['THETA(3)'] == Parameter('THETA(3)', 0.25, lower=-0.99, upper=1000000)
    assert model.parameters['OMEGA(1,1)'] == Parameter('OMEGA(1,1)', 0.1, lower=0, upper=sympy.oo)
    assert model.parameters['OMEGA(2,2)'] == Parameter('OMEGA(2,2)', 0.2, lower=0, upper=sympy.oo)
    assert model.parameters['SIGMA(1,1)'] == Parameter('SIGMA(1,1)', 0.3, lower=0, upper=sympy.oo)
    model.update_source()
    thetas = model.control_stream.get_records('THETA')
    assert str(thetas[0]) == '$THETA (0,0.75) ; PTVCL\n'
    assert str(thetas[1]) == '$THETA (0,0.5) ; PTVV\n'
    assert str(thetas[2]) == '$THETA (-.99,0.25)\n'
    omegas = model.control_stream.get_records('OMEGA')
    assert str(omegas[0]) == '$OMEGA DIAGONAL(2)\n 0.1  ;       IVCL\n 0.2  ;        IVV\n\n'
    sigmas = model.control_stream.get_records('SIGMA')
    assert str(sigmas[0]) == '$SIGMA 0.3\n'

    model = Model(pheno_path)
    params = model.parameters
    params['THETA(1)'].init = 18
    model.parameters = params
    assert model.parameters['THETA(1)'] == Parameter('THETA(1)', 18, lower=0, upper=1000000)
    assert model.parameters['THETA(2)'] == Parameter('THETA(2)', 1.00916, lower=0, upper=1000000)


def test_adjust_iovs(testdata):
    model = Model(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'multEST' / 'noSIM' / 'withBayes.mod'
    )
    model.parameters
    rvs = model.random_variables

    assert rvs[0].level == 'IIV'
    assert rvs[3].level == 'IOV'
    assert rvs[4].level == 'IOV'
    assert rvs[6].level == 'IOV'

    model = Model(testdata / 'nonmem' / 'qa' / 'iov.mod')
    rvs = model.random_variables
    assert rvs[0].level == 'IIV'
    assert rvs[1].level == 'IIV'
    assert rvs[2].level == 'IOV'
    assert rvs[3].level == 'IOV'
    assert rvs[4].level == 'IOV'
    assert rvs[5].level == 'IOV'


@pytest.mark.parametrize(
    'param_new,init_expected,buf_new',
    [
        (Parameter('COVEFF', 0.2), 0.2, '$THETA  0.2 ; COVEFF'),
        (Parameter('THETA', 0.1), 0.1, '$THETA  0.1 ; THETA'),
        (Parameter('THETA', 0.1, 0, fix=True), 0.1, '$THETA  (0,0.1) FIX ; THETA'),
        (Parameter('RUV_prop', 0.1), 0.1, '$THETA  0.1 ; RUV_prop'),
    ],
)
def test_add_parameters(pheno_path, param_new, init_expected, buf_new):
    model = Model(pheno_path)
    pset = model.parameters

    assert len(pset) == 6

    pset.append(param_new)
    model.parameters = pset

    assert len(pset) == 7
    assert model.parameters[param_new.name].init == init_expected

    rec_ref = (
        f'$THETA (0,0.00469307) ; PTVCL\n'
        f'$THETA (0,1.00916) ; PTVV\n'
        f'$THETA (-.99,.1)\n'
        f'{buf_new}\n'
    )

    model.update_source()
    rec_mod = ''
    for rec in model.control_stream.get_records('THETA'):
        rec_mod += str(rec)

    assert rec_ref == rec_mod


def test_add_two_parameters(pheno_path):
    model = Model(pheno_path)
    pset = model.parameters

    assert len(pset) == 6

    param_1 = Parameter('COVEFF1', 0.2)
    param_2 = Parameter('COVEFF2', 0.1)
    pset.append(param_1)
    pset.append(param_2)
    model.parameters = pset
    model.update_source()

    assert len(pset) == 8
    assert model.parameters[param_1.name].init == 0.2
    assert model.parameters[param_2.name].init == 0.1


@pytest.mark.parametrize(
    'statement_new,buf_new',
    [
        (Assignment(S('CL'), sympy.Integer(2)), 'CL = 2'),
        (Assignment(S('Y'), S('THETA(4)') + S('THETA(5)')), 'Y = THETA(4) + THETA(5)'),
    ],
)
def test_add_statements(pheno_path, statement_new, buf_new):
    model = Model(pheno_path)
    sset = model.statements
    assert len(sset) == 15

    # Insert new statement before ODE system.
    new_sset = ModelStatements()
    for s in sset:
        if isinstance(s, ODESystem):
            new_sset.append(statement_new)
        new_sset.append(s)

    model.statements = new_sset
    model.update_source()

    assert len(model.statements) == 16

    parser = NMTranParser()
    stream = parser.parse(model.model_code)

    assert str(model.control_stream) == str(stream)

    rec_ref = (
        f'$PK\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'CL=TVCL*EXP(ETA(1))\n'
        f'V=TVV*EXP(ETA(2))\n'
        f'S1=V\n'
        f'{buf_new}\n\n'
    )

    rec_mod = str(model.control_stream.get_records('PK')[0])

    assert rec_ref == rec_mod


@pytest.mark.parametrize(
    'param_new, statement_new, buf_new',
    [
        (Parameter('X', 0.1), Assignment(S('Y'), S('X') + S('S1')), 'Y = S1 + THETA(4)'),
    ],
)
def test_add_parameters_and_statements(pheno_path, param_new, statement_new, buf_new):
    model = Model(pheno_path)

    pset = model.parameters
    pset.append(param_new)
    model.parameters = pset

    sset = model.statements

    # Insert new statement before ODE system.
    new_sset = ModelStatements()
    for s in sset:
        if isinstance(s, ODESystem):
            new_sset.append(statement_new)
        new_sset.append(s)

    model.statements = new_sset
    model.update_source()

    rec = (
        f'$PK\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'CL=TVCL*EXP(ETA(1))\n'
        f'V=TVV*EXP(ETA(2))\n'
        f'S1=V\n'
        f'{buf_new}\n\n'
    )

    assert str(model.get_pred_pk_record()) == rec


@pytest.mark.parametrize('rv_new, buf_new', [(Parameter('omega', 0.1), '$OMEGA  0.1')])
def test_add_random_variables(pheno_path, rv_new, buf_new):
    model = Model(pheno_path)
    rvs = model.random_variables
    pset = model.parameters

    eta = RandomVariable.normal('eta_new', 'iiv', 0, S(rv_new.name))

    rvs.append(eta)
    pset.append(rv_new)

    model.random_variables = rvs
    model.parameters = pset

    model.update_source()

    rec_ref = (
        f'$OMEGA DIAGONAL(2)\n'
        f' 0.0309626  ;       IVCL\n'
        f' 0.031128  ;        IVV\n\n'
        f'{buf_new} ; omega\n'
    )

    rec_mod = ''
    for rec in model.control_stream.get_records('OMEGA'):
        rec_mod += str(rec)

    assert rec_mod == rec_ref

    rv = model.random_variables['eta_new']

    assert rv.sympy_rv.pspace.distribution.mean == 0
    assert (rv.sympy_rv.pspace.distribution.std ** 2).name == 'omega'


def test_add_random_variables_and_statements(pheno_path):
    model = Model(pheno_path)

    rvs = model.random_variables
    pset = model.parameters

    eta = RandomVariable.normal('ETA_NEW', 'iiv', 0, S('omega'))
    rvs.append(eta)
    pset.append(Parameter('omega', 0.1))

    eps = RandomVariable.normal('EPS_NEW', 'ruv', 0, S('sigma'))
    rvs.append(eps)
    pset.append(Parameter('sigma', 0.1))

    model.random_variables = rvs
    model.parameters = pset

    sset = model.get_pred_pk_record().statements

    statement_new = Assignment(S('X'), 1 + S(eps.name) + S(eta.name))
    sset.append(statement_new)
    model.get_pred_pk_record().statements = sset

    model.update_source()

    assert str(model.get_pred_pk_record()).endswith('X = 1 + ETA(3) + EPS(2)\n\n')


def test_results(pheno_path):
    model = Model(pheno_path)
    assert len(model.modelfit_results) == 1  # A chain of one estimation


def test_minimal(datadir):
    path = datadir / 'minimal.mod'
    model = Model(path)
    assert len(model.statements) == 1
    assert model.statements[0].expression == symbol('THETA(1)') + symbol('ETA(1)') + symbol(
        'EPS(1)'
    )


def test_copy(datadir):
    path = datadir / 'minimal.mod'
    model = Model(path)
    copy = model.copy()
    assert id(model) != id(copy)
    assert model.statements[0].expression == symbol('THETA(1)') + symbol('ETA(1)') + symbol(
        'EPS(1)'
    )


def test_initial_individual_estimates(datadir):
    path = datadir / 'minimal.mod'
    model = Model(path)
    assert model.initial_individual_estimates is None

    path = datadir / 'pheno_etas.mod'
    model = Model(path)
    inits = model.initial_individual_estimates
    assert len(inits) == 59
    assert len(inits.columns) == 2
    assert inits['ETA(1)'][2] == -0.166321


def test_update_individual_estimates(datadir):
    with Patcher(additional_skip_names=['pkgutil']) as patcher:
        fs = patcher.fs
        fs.add_real_file(datadir / 'pheno_real.mod', target_path='run1.mod')
        fs.add_real_file(datadir / 'pheno_real.phi', target_path='run1.phi')
        fs.add_real_file(datadir / 'pheno_real.lst', target_path='run1.lst')
        fs.add_real_file(datadir / 'pheno_real.ext', target_path='run1.ext')
        fs.add_real_file(datadir / 'pheno.dta', target_path='pheno.dta')
        model = Model('run1.mod')
        model2 = model.copy()
        model2.name = 'run2'
        model2.update_individual_estimates(model)
        model2.update_source()
        with open('run2_input.phi', 'r') as fp, open('run1.phi') as op:
            assert fp.read() == op.read()
        assert model2.model_code.endswith(
            """$ESTIMATION METHOD=1 INTERACTION MCETA=1
$COVARIANCE UNCONDITIONAL
$TABLE ID TIME DV AMT WGT APGR IPRED PRED RES TAD CWRES NPDE NOAPPEND
       NOPRINT ONEHEADER FILE=sdtab2
$ETAS FILE=run2_input.phi"""
        )


@pytest.mark.parametrize(
    'buf_new, len_expected',
    [
        (
            'IF(AMT.GT.0) BTIME=TIME\nTAD=TIME-BTIME\n'
            'TVCL=THETA(1)*WGT\nTVV=THETA(2)*WGT\n'
            'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\nCL=TVCL*EXP(ETA(1))'
            '\nV=TVV*EXP(ETA(2))\nS1=V\nY=A+B',
            9,
        ),
        (
            'IF(AMT.GT.0) BTIME=TIME\nTAD=TIME-BTIME\n'
            'TVCL=THETA(1)*WGT\nTVV=THETA(2)*WGT\n'
            'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\nCL=TVCL*EXP(ETA(1))'
            '\nV=TVV*EXP(ETA(2))\nS1=2*V',
            8,
        ),
    ],
)
def test_statements_setter(pheno_path, buf_new, len_expected):
    model = Model(pheno_path)

    parser = NMTranParser()
    statements_new = parser.parse(f'$PRED\n{buf_new}').records[0].statements

    assert len(model.statements) == 15
    assert len(statements_new) == len_expected

    model.statements = statements_new

    assert len(model.statements) == len_expected
    assert model.statements == statements_new


def test_deterministic_theta_comments(pheno_path):
    model = Model(pheno_path)

    no_option = 0
    for theta_record in model.control_stream.get_records('THETA'):
        no_option += len(theta_record.root.all('option'))

    assert no_option == 0


def test_remove_eta(pheno_path):
    model = Model(pheno_path)
    rvs = model.random_variables
    eta1 = rvs['ETA(1)']
    del rvs[eta1]
    model.update_source()
    assert model.model_code.split('\n')[12] == 'V = TVV*EXP(ETA(1))'


def test_symbol_names_in_comment(pheno_path):
    with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
        model = Model(pheno_path)
        assert model.statements[2].expression == S('PTVCL') * S('WGT')

        code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
TV = THETA(1)
Y = TV + ETA(1) + ERR(1)

$THETA 0.1  ; TV
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
        model = Model(StringIO(code))
        with pytest.warns(UserWarning):
            assert model.parameters.names == ['THETA(1)', 'OMEGA(1,1)', 'SIGMA(1,1)']


def test_symbol_names_in_abbr(testdata):
    with ConfigurationContext(conf, parameter_names=['abbr', 'basic']):
        model = Model(testdata / 'nonmem' / 'pheno_abbr.mod')
        pset, rvs = model.parameters, model.random_variables

        assert 'THETA_CL' in pset.names
        assert 'ETA_CL' in [eta.name for eta in rvs.etas]


@pytest.mark.parametrize(
    'parameter_names, assignments, params, etas',
    [
        (
            ['abbr', 'comment', 'basic'],
            [
                Assignment(S('CL'), S('THETA_CL') * sympy.exp(S('ETA_CL'))),
                Assignment(S('V'), S('TVV') * sympy.exp(S('ETA(2)'))),
            ],
            ['THETA_CL', 'TVV', 'IVCL', 'OMEGA(2,2)'],
            ['ETA_CL', 'ETA(2)'],
        ),
        (
            ['comment', 'abbr', 'basic'],
            [
                Assignment(S('CL'), S('TVCL') * sympy.exp(S('ETA_CL'))),
                Assignment(S('V'), S('TVV') * sympy.exp(S('ETA(2)'))),
            ],
            ['TVCL', 'TVV', 'IVCL', 'OMEGA(2,2)'],
            ['ETA_CL', 'ETA(2)'],
        ),
        (
            ['abbr', 'basic'],
            [
                Assignment(S('CL'), S('THETA_CL') * sympy.exp(S('ETA_CL'))),
                Assignment(S('V'), S('THETA(2)') * sympy.exp(S('ETA(2)'))),
            ],
            ['THETA_CL', 'THETA(2)', 'OMEGA(1,1)', 'OMEGA(2,2)'],
            ['ETA_CL', 'ETA(2)'],
        ),
        (
            ['basic'],
            [
                Assignment(S('CL'), S('THETA(1)') * sympy.exp(S('ETA(1)'))),
                Assignment(S('V'), S('THETA(2)') * sympy.exp(S('ETA(2)'))),
            ],
            ['THETA(1)', 'THETA(2)', 'OMEGA(1,1)', 'OMEGA(2,2)'],
            ['ETA(1)', 'ETA(2)'],
        ),
        (
            ['basic', 'comment'],
            [
                Assignment(S('CL'), S('THETA(1)') * sympy.exp(S('ETA(1)'))),
                Assignment(S('V'), S('THETA(2)') * sympy.exp(S('ETA(2)'))),
            ],
            ['THETA(1)', 'THETA(2)', 'OMEGA(1,1)', 'OMEGA(2,2)'],
            ['ETA(1)', 'ETA(2)'],
        ),
        (
            ['basic', 'abbr'],
            [
                Assignment(S('CL'), S('THETA(1)') * sympy.exp(S('ETA(1)'))),
                Assignment(S('V'), S('THETA(2)') * sympy.exp(S('ETA(2)'))),
            ],
            ['THETA(1)', 'THETA(2)', 'OMEGA(1,1)', 'OMEGA(2,2)'],
            ['ETA(1)', 'ETA(2)'],
        ),
    ],
)
def test_symbol_names_priority(testdata, parameter_names, assignments, params, etas):
    with ConfigurationContext(conf, parameter_names=parameter_names):
        model = Model(testdata / 'nonmem' / 'pheno_abbr_comments.mod')
        sset, pset, rvs = model.statements, model.parameters, model.random_variables

        assert all(str(a) in [str(s) for s in sset] for a in assignments)
        assert all(p in pset.names for p in params)
        assert all(eta in [rv.name for rv in rvs] for eta in etas)


def test_clashing_parameter_names(datadir):
    with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
        model = Model(datadir / 'pheno_clashing_symbols.mod')
        with pytest.warns(UserWarning):
            model.statements
        assert model.parameters.names == ['THETA(1)', 'TVV', 'IVCL', 'OMEGA(2,2)', 'SIGMA(1,1)']

        code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1  ; TV
$OMEGA 0.01 ; TV
$SIGMA 1 ; TV
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
        model = Model(StringIO(code))
        with pytest.warns(UserWarning):
            assert model.parameters.names == ['TV', 'OMEGA(1,1)', 'SIGMA(1,1)']

        code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + THETA(2)

$THETA 0.1  ; TV
$THETA 0.1  ; TV
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
        model = Model(StringIO(code))
        with pytest.warns(UserWarning):
            assert model.parameters.names == ['TV', 'THETA(2)']


def test_missing_parameter_names_settings(pheno_path):
    with ConfigurationContext(conf, parameter_names=['comment']):
        model = Model(pheno_path)
        with pytest.raises(ValueError):
            model.statements


def test_abbr_write(pheno_path):
    with ConfigurationContext(conf, write_etas_in_abbr=True):
        model = Model(pheno_path)
        add_iiv(model, 'S1', 'add')
        model.update_source()

        assert 'ETA(S1)' in model.model_code
        assert 'ETA_S1' in [rv.name for rv in model.random_variables]
        assert S('ETA_S1') in model.statements.free_symbols

        model.update_source()

        assert 'ETA(S1)' in model.model_code
        assert 'ETA_S1' in [rv.name for rv in model.random_variables]
        assert S('ETA_S1') in model.statements.free_symbols

        model = Model(pheno_path)
        add_iiv(model, 'S1', 'add', eta_names='new_name')

        with pytest.warns(UserWarning, match='Not valid format of name new_name'):
            model.update_source()
            assert 'ETA(3)' in model.model_code


def test_abbr_read_write(pheno_path):
    with ConfigurationContext(
        conf, parameter_names=['abbr', 'comment', 'basic'], write_etas_in_abbr=True
    ):
        model_write = Model(pheno_path)
        add_iiv(model_write, 'S1', 'add')
        model_read = Model(StringIO(model_write.model_code))
        assert model_read.model_code == model_write.model_code
        assert model_read.statements == model_write.statements
        assert not (
            model_read.random_variables - model_write.random_variables
        )  # Different order due to renaming in read


def test_dv_symbol(pheno_path):
    model = Model(pheno_path)
    assert model.dependent_variable.name == 'Y'


def test_insert_unknown_record(pheno_path):
    model = Model(pheno_path)
    model.control_stream.insert_record('$TRIREME one')
    assert model.model_code.split('\n')[-1] == '$TRIREME one'

    model.control_stream.insert_record('\n$OA two')
    assert model.model_code.split('\n')[-1] == '$OA two'


def test_parse_illegal_record_name():
    parser = NMTranParser()
    with pytest.raises(ModelSyntaxError):
        parser.parse('$PROBLEM MYPROB\n$1REC\n')


def test_frem_model():
    code = """$PROBLEM    MOXONIDINE PK,FINAL ESTIMATES,simulated data
$INPUT      ID VISI DROP DGRP DOSE DROP DROP DROP DROP NEUY SCR AGE
            SEX DROP WT DROP ACE DIG DIU DROP TAD TIME DROP CRCL AMT
            SS II DROP DROP DROP DV DROP DROP MDV FREMTYPE
$DATA      ../frem_dataset.dta IGNORE=@
$SUBROUTINE ADVAN2 TRANS1
$OMEGA  BLOCK(1)
 1.47E-06  ;     IOV_CL
$OMEGA  BLOCK(1) SAME
$OMEGA  BLOCK(1)
 5.06E-05  ;     IOV_KA
$OMEGA  BLOCK(1) SAME
$OMEGA  BLOCK(5)
 0.416046
 0.389578 0.578819  ;   IIV_CL_V
 0.00327757 0.00386591 0.258203  ;     IIV_KA
 -0.0809376 -0.130036 0.0777389 1  ;    BSV_AGE
 0.0880473 0.0821262 -0.0578989 -0.11524 1  ;    BSV_SEX
$PK
   VIS3               = 0
   IF(VISI.EQ.3) VIS3 = 1
   VIS8               = 0
   IF(VISI.EQ.8) VIS8 = 1
   KPCL  = VIS3*ETA(1)+VIS8*ETA(2)
   KPKA  = VIS3*ETA(3)+VIS8*ETA(4)
   TVCL  = THETA(1)
   TVV   = THETA(2)
   CL    = TVCL*EXP(ETA(5)+KPCL)
   V     = TVV*EXP(ETA(6))
   KA    = THETA(3)*EXP(ETA(7)+KPKA)
   ALAG1 = THETA(4)
   K     = CL/V
   S2    = V
    SDC8 = 7.82226906804
    SDC9 = 0.404756978659

$ERROR
     IPRED = LOG(.025)
     WA     = 1
     W      = WA
     IF(F.GT.0) IPRED = LOG(F)
     IRES  = IPRED-DV
     IWRES = IRES/W
     Y     = IPRED+ERR(1)*W

;;;FREM CODE BEGIN COMPACT
;;;DO NOT MODIFY
    IF (FREMTYPE.EQ.100) THEN
;      AGE  7.82226906804
       Y = THETA(5) + ETA(8)*SDC8 + EPS(2)
       IPRED = THETA(5) + ETA(8)*SDC8
    END IF
    IF (FREMTYPE.EQ.200) THEN
;      SEX  0.404756978659
       Y = THETA(6) + ETA(9)*SDC9 + EPS(2)
       IPRED = THETA(6) + ETA(9)*SDC9
    END IF
;;;FREM CODE END COMPACT
$THETA  (0,32.901) ; POP_TVCL
$THETA  (0,115.36) ; POP_TVV
$THETA  (0,1.45697) ; POP_TVKA
$THETA  (0,0.0818029) ; POP_LAG
$THETA  65.1756756757 FIX ; TV_AGE
 1.2027027027 FIX ; TV_SEX
$SIGMA  0.112373
$SIGMA  0.0000001  FIX  ;     EPSCOV
$ESTIMATION METHOD=1 MAXEVAL=9999 NONINFETA=1 MCETA=1
"""
    model = Model(StringIO(code))
    rvs = model.random_variables
    assert len(rvs) == 11


@pytest.mark.parametrize(
    'model_path, transformation',
    [
        ('nonmem/pheno.mod', set_zero_order_elimination),
    ],
)
def test_des(testdata, model_path, transformation):
    model_ref = Model(testdata / model_path)
    transformation(model_ref)

    model_des = Model(StringIO(model_ref.model_code))

    assert model_ref.statements.ode_system == model_des.statements.ode_system


def test_cmt_warning(testdata):
    model_original = Model(testdata / 'nonmem' / 'models' / 'mox1.mod')

    model_str = model_original.model_code.replace('CMT=DROP', 'CMT')
    model = Model(StringIO(model_str))
    model.dataset_path = model_original.dataset_path

    set_zero_order_absorption(model)

    with pytest.raises(UserWarning, match='Compartment structure has been updated'):
        model.update_source(nofiles=True)


@pytest.mark.parametrize(
    'estcode,correct',
    [
        ('$ESTIMATION METH=COND INTERACTION', [EstimationMethod('foce', interaction=True)]),
        ('$ESTIMATION INTER METH=COND', [EstimationMethod('foce', interaction=True)]),
        ('$ESTM METH=1 INTERACTION', [EstimationMethod('foce', interaction=True)]),
        ('$ESTIM METH=1', [EstimationMethod('foce', interaction=False)]),
        ('$ESTIMA METH=0', [EstimationMethod('fo', interaction=False)]),
        ('$ESTIMA METH=ZERO', [EstimationMethod('fo', interaction=False)]),
        ('$ESTIMA INTER', [EstimationMethod('fo', interaction=True)]),
        ('$ESTIMA INTER\n$COV', [EstimationMethod('fo', interaction=True, cov=True)]),
        (
            '$ESTIMA METH=COND INTER\n$EST METH=COND',
            [
                EstimationMethod('foce', interaction=True),
                EstimationMethod('foce', interaction=False),
            ],
        ),
        ('$ESTIMATION METH=SAEM', [EstimationMethod('saem', interaction=False)]),
        ('$ESTIMATION METH=1 LAPLACE', [EstimationMethod('foce', interaction=False, laplace=True)]),
        (
            '$ESTIMATION METH=0 MAXEVAL=0',
            [EstimationMethod('fo', interaction=False, evaluation=True)],
        ),
        (
            '$ESTIMATION METH=IMP EONLY=1',
            [EstimationMethod('imp', interaction=False, evaluation=True)],
        ),
        (
            '$ESTIMATION METH=COND MAXEVAL=9999',
            [EstimationMethod('foce', maximum_evaluations=9999)],
        ),
    ],
)
def test_estimation_steps_getter(estcode, correct):
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    code += estcode
    model = Model(StringIO(code))
    assert model.estimation_steps == correct


def test_estimation_steps_getter_options():
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 SADDLE_RESET=1
'''
    model = Model(StringIO(code))
    assert model.estimation_steps[0].method == 'FOCE'
    assert model.estimation_steps[0].tool_options['SADDLE_RESET'] == '1'


@pytest.mark.parametrize(
    'estcode,kwargs,rec_ref',
    [
        ('$EST METH=COND INTER', {'method': 'fo'}, '$ESTIMATION METHOD=ZERO INTER'),
        ('$EST METH=COND INTER', {'interaction': False}, '$ESTIMATION METHOD=COND'),
        ('$EST METH=COND INTER', {'cov': True}, '$COVARIANCE'),
        (
            '$EST METH=COND INTER MAXEVAL=99999',
            {'method': 'fo'},
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=99999',
        ),
        (
            '$EST METH=COND INTER POSTHOC',
            {'method': 'fo'},
            '$ESTIMATION METHOD=ZERO INTER POSTHOC',
        ),
        ('$EST METH=COND INTER', {'laplace': True}, '$ESTIMATION METHOD=COND LAPLACE INTER'),
    ],
)
def test_estimation_steps_setter(estcode, kwargs, rec_ref):
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    code += estcode
    model = Model(StringIO(code))
    for key, value in kwargs.items():
        setattr(model.estimation_steps[0], key, value)
    assert model.model_code.split('\n')[-2] == rec_ref


@pytest.mark.parametrize(
    'estcode,kwargs,error_msg',
    [
        (
            '$EST METH=COND MAXEVAL=0',
            {'tool_options': {'MAXEVAL': 999}},
            'MAXEVAL already set as attribute in estimation method object',
        ),
        (
            '$EST METH=COND INTER',
            {'tool_options': {'INTERACTION': None}},
            'INTERACTION already set as attribute in estimation method object',
        ),
    ],
)
def test_set_estimation_steps_option_clash(estcode, kwargs, error_msg):
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    code += estcode
    model = Model(StringIO(code))

    for key, value in kwargs.items():
        setattr(model.estimation_steps[0], key, value)

    with pytest.raises(ValueError) as excinfo:
        model.update_source(nofiles=True)
    assert error_msg == str(excinfo.value)


def test_add_estimation_step():
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$EST METH=COND INTER
'''
    model = Model(StringIO(code))
    est_new = EstimationMethod('IMP', interaction=True, tool_options={'saddle_reset': 1})
    model.estimation_steps.append(est_new)
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=IMP INTER SADDLE_RESET=1'
    est_new = EstimationMethod('SAEM', interaction=True)
    model.estimation_steps.insert(0, est_new)
    assert model.model_code.split('\n')[-4] == '$ESTIMATION METHOD=SAEM INTER'
    est_new = EstimationMethod('FO', evaluation=True)
    model.estimation_steps.append(est_new)
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=ZERO MAXEVAL=0'
    est_new = EstimationMethod('IMP', evaluation=True)
    model.estimation_steps.append(est_new)
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=IMP EONLY=1'


def test_remove_estimation_step():
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$EST METH=COND INTER
'''
    model = Model(StringIO(code))
    del model.estimation_steps[0]
    assert not model.estimation_steps
    model.update_source()
    assert model.model_code.split('\n')[-2] == '$SIGMA 1'


def test_update_source_comments():
    code = """
$PROBLEM    run3.mod PHENOBARB SIMPLE MODEL
$DATA      ../../pheno.dta IGNORE=@
$INPUT      ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK


IF(AMT.GT.0) BTIME=TIME
TAD=TIME-BTIME
      TVCL=THETA(1)*WGT
      TVV=THETA(2)*WGT
IF(APGR.LT.5) TVV=TVV*(1+THETA(3))
      CL=TVCL*EXP(ETA(1))
      V=TVV*EXP(ETA(2))
      S1=V
$ERROR

      W=F
      Y=F+W*EPS(1)

      IPRED=F         ;  individual-specific prediction
      IRES=DV-IPRED   ;  individual-specific residual
      IWRES=IRES/W    ;  individual-specific weighted residual

$THETA  (0,0.00469555) ; CL
$THETA  (0,0.984258) ; V
$THETA  (-.99,0.15892)
$OMEGA  DIAGONAL(2)
 0.0293508  ;       IVCL
 0.027906  ;        IVV
$SIGMA  0.013241
$ESTIMATION METHOD=1 INTERACTION MAXEVALS=9999
$COVARIANCE UNCONDITIONAL
$TABLE      ID TIME AMT WGT APGR IPRED PRED TAD CWRES NPDE NOAPPEND
            NOPRINT ONEHEADER FILE=mytab3
"""
    with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
        with pytest.warns(UserWarning):
            model = Model(StringIO(code))
            model.update_source()


def test_convert_model(testdata):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 2 ; OM1
$SIGMA 3 ; SI1
$ESTIMATION METHOD=1 INTER
"""
    base = Model(StringIO(code))
    base.dataset_path = testdata / 'nonmem' / 'file.csv'
    model = convert_model(base)
    model.dataset_path = "file.csv"  # Otherwise we get full path
    correct = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 2 ; OM1
$SIGMA 3 ; SI1
$ESTIMATION METHOD=1 INTER
"""
    assert model.model_code == correct



"""
:meta private:
"""

import numpy as np
import sympy

from pharmpy import math
from pharmpy.modeling.help_functions import _get_etas
from pharmpy.parameter import Parameter
from pharmpy.random_variables import RandomVariables


def create_joint_distribution(model, rvs=None):
    """
    Combines some or all etas into a joint distribution.

    The etas must be IIVs and cannot
    be fixed. Initial estimates for covariance between the etas is dependent on whether
    the model has results from a previous results. In that case, the correlation will
    be calculated from individual estimates, otherwise correlation will be set to 10%.

    Parameters
    ----------
    model : Model
        Pharmpy model
    rvs : list
        Sequence of etas or names of etas to combine. If None, all etas that are IIVs and
        non-fixed will be used (full block). None is default.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_joint_distribution
    >>> model = load_example_model("pheno")
    >>> model.random_variables.etas
    ETA(1) ~ 𝒩 (0, OMEGA(1,1))
    ETA(2) ~ 𝒩 (0, OMEGA(2,2))
    >>> create_joint_distribution(model, ['ETA(1)', 'ETA(2)'])      # doctest: +ELLIPSIS
    <...>
    >>> model.random_variables.etas
    ⎡ETA(1)⎤     ⎧⎡0⎤  ⎡ OMEGA(1,1)   IIV_CL_IIV_V⎤⎫
    ⎢      ⎥ ~ 𝒩 ⎪⎢ ⎥, ⎢                          ⎥⎪
    ⎣ETA(2)⎦     ⎩⎣0⎦  ⎣IIV_CL_IIV_V   OMEGA(2,2) ⎦⎭

    See also
    --------
    split_joint_distribution : split etas into separate distributions
    """
    all_rvs = model.random_variables
    if rvs is None:
        iiv_rvs = model.random_variables.iiv
        nonfix = RandomVariables()
        for rv in iiv_rvs:
            for name in iiv_rvs.parameter_names:
                if model.parameters[name].fix:
                    break
            else:
                nonfix.append(rv)
        rvs = nonfix.names
    else:
        for name in rvs:
            if name in all_rvs and all_rvs[name].level == 'IOV':
                raise ValueError(
                    f'{name} describes IOV: Joining IOV random variables is currently not supported'
                )
    if len(rvs) == 1:
        raise ValueError('At least two random variables are needed')

    sset = model.statements
    paramnames = []
    for rv in rvs:
        parameter_names = '_'.join(
            [s.symbol.name for s in sset if sympy.Symbol(rv) in s.rhs_symbols]
        )
        paramnames.append(parameter_names)

    cov_to_params = all_rvs.join(rvs, name_template='IIV_{}_IIV_{}', param_names=paramnames)

    pset = model.parameters
    for cov_name, param_names in cov_to_params.items():
        parent_params = (pset[param_names[0]], pset[param_names[1]])
        covariance_init = _choose_param_init(model, all_rvs, parent_params)
        param_new = Parameter(cov_name, covariance_init)
        pset.append(param_new)

    return model


def split_joint_distribution(model, rvs=None):
    """
    Splits etas following a joint distribution into separate distributions.

    Parameters
    ----------
    model : Model
        Pharmpy model
    rvs : str, list
        Name/names of etas to separate. If None, all etas that are IIVs and
        non-fixed will become single. None is default.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> create_joint_distribution(model, ['ETA(1)', 'ETA(2)'])      # doctest: +ELLIPSIS
    <...>
    >>> model.random_variables.etas
    ⎡ETA(1)⎤     ⎧⎡0⎤  ⎡ OMEGA(1,1)   IIV_CL_IIV_V⎤⎫
    ⎢      ⎥ ~ 𝒩 ⎪⎢ ⎥, ⎢                          ⎥⎪
    ⎣ETA(2)⎦     ⎩⎣0⎦  ⎣IIV_CL_IIV_V   OMEGA(2,2) ⎦⎭
    >>> split_joint_distribution(model, ['ETA(1)', 'ETA(2)'])       # doctest: +ELLIPSIS
    <...>
    >>> model.random_variables.etas
    ETA(1) ~ 𝒩 (0, OMEGA(1,1))
    ETA(2) ~ 𝒩 (0, OMEGA(2,2))

    See also
    --------
    create_joint_distribution : combine etas into a join distribution
    """
    all_rvs = model.random_variables
    rvs = _get_etas(model, rvs)

    parameters_before = all_rvs.parameter_names
    all_rvs.unjoin(rvs)
    parameters_after = all_rvs.parameter_names

    removed_parameters = set(parameters_before) - set(parameters_after)
    pset = model.parameters
    for param in removed_parameters:
        del pset[param]

    return model


def _choose_param_init(model, rvs, params):
    res = model.modelfit_results
    rvs_names = [rv.name for rv in rvs]

    etas = []
    for i in range(len(rvs)):
        elem = rvs.covariance_matrix.row(i).col(i)[0]
        if str(elem) in [p.name for p in params]:
            etas.append(rvs_names[i])

    sd = np.array([np.sqrt(params[0].init), np.sqrt(params[1].init)])
    init_default = round(0.1 * sd[0] * sd[1], 7)

    if res is not None:
        try:
            ie = res.individual_estimates
            if not all(eta in ie.columns for eta in etas):
                return init_default
        except KeyError:
            return init_default
        eta_corr = ie[etas].corr()
        cov = math.corr2cov(eta_corr.to_numpy(), sd)
        cov[cov == 0] = 0.0001
        cov = math.nearest_posdef(cov)
        return round(cov[1][0], 7)
    else:
        return init_default



import os
import os.path

import pytest

from pharmpy.workflows.database import ModelDatabase
from pharmpy.workflows.databases.local_directory import LocalDirectoryDatabase
from pharmpy.workflows.databases.null_database import NullModelDatabase, NullToolDatabase


def test_base_class():
    db = ModelDatabase()
    with pytest.raises(NotImplementedError):
        db.store_local_file(None, "file.txt")


def test_local_directory(fs):
    os.mkdir("database")
    db = LocalDirectoryDatabase("database")
    with open("file.txt", "w") as fh:
        print("Hello!", file=fh)
    db.store_local_file(None, "file.txt")
    with open("database/file.txt", "r") as fh:
        assert fh.read() == "Hello!\n"

    dirname = "doesnotexist"
    db = LocalDirectoryDatabase(dirname)
    assert os.path.isdir(dirname)


def test_null_database():
    db = NullToolDatabase("any", sl1=23, model=45, opr=12, dummy="some dummy kwargs")
    db.store_local_file("path")
    db = NullModelDatabase(klr=123, f="oe")
    db.store_local_file("path", 34)



import pytest
import sympy

import pharmpy.symbols
from pharmpy import Model
from pharmpy.statements import Assignment, CompartmentalSystem, ModelStatements, ODESystem


def S(x):
    return pharmpy.symbols.symbol(x)


def test_str(testdata):
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    assert str(s1) == 'KA := X + Y'
    s2 = Assignment(S('X2'), sympy.exp('X'))
    a = str(s2).split('\n')
    assert a[0].startswith(' ')
    assert len(a) == 2

    model = Model(testdata / 'nonmem' / 'pheno.mod')
    assert 'THETA(2)' in str(model.statements)
    assert 'THETA(2)' in repr(model.statements)


def test_subs(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements

    statements.subs({'ETA(1)': 'ETAT1'})

    assert statements[5].expression == S('TVCL') * sympy.exp(S('ETAT1'))

    statements.subs({'TVCL': 'TVCLI'})

    assert statements[2].symbol == S('TVCLI')
    assert statements[5].expression == S('TVCLI') * sympy.exp(S('ETAT1'))

    assert statements.ode_system.free_symbols == {S('V'), S('CL'), S('AMT'), S('t')}

    statements.subs({'V': 'V2'})

    assert statements.ode_system.free_symbols == {S('CL'), S('AMT'), S('t'), S('V2')}


def test_ode_system_base_class():
    odes = ODESystem()
    assert odes.free_symbols == set()
    assert odes.rhs_symbols == set()
    odes.subs({})
    assert odes == odes
    assert odes != CompartmentalSystem()
    assert str(odes) == 'ODESystem()'


def test_ode_free_symbols(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')

    assert model.statements.ode_system.free_symbols == {S('V'), S('CL'), S('AMT'), S('t')}


def test_find_assignment(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements

    assert str(statements.find_assignment('CL').expression) == 'TVCL*exp(ETA(1))'
    assert str(statements.find_assignment('S1').expression) == 'V'

    statements.append(Assignment(S('CL'), S('TVCL') + S('V')))

    assert str(statements.find_assignment('CL').expression) == 'TVCL + V'


def test_eq_assignment(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements
    statement_btime = statements[0]
    statement_tad = statements[1]

    assert statement_btime == statement_btime
    assert statement_btime != statement_tad


def test_eq_modelstatements(testdata):
    model_min = Model(testdata / 'nonmem' / 'minimal.mod')
    model_pheno = Model(testdata / 'nonmem' / 'pheno_real.mod')

    assert model_min.statements == model_min.statements
    assert model_pheno.statements == model_pheno.statements
    assert model_min.statements != model_pheno.statements


def test_remove_symbol_definition():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    s2 = Assignment(S('Z'), sympy.Integer(23) + S('M'))
    s3 = Assignment(S('M'), sympy.Integer(2))
    s4 = Assignment(S('G'), sympy.Integer(3))
    s = ModelStatements([s4, s3, s2, s1])
    s.remove_symbol_definitions([S('Z')], s1)
    assert s == ModelStatements([s4, s1])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X') + S('K'))
    s = ModelStatements([s1, s2, s3, s4])
    s.remove_symbol_definitions([S('CL')], s4)
    assert s == ModelStatements([s1, s4])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X') + S('K'))
    s5 = Assignment(S('KR'), S('CL'))
    s = ModelStatements([s1, s2, s3, s4, s5])
    s.remove_symbol_definitions([S('CL')], s4)
    assert s == ModelStatements([s1, s2, s3, s4, s5])

    s1 = Assignment(S('K'), sympy.Integer(16))
    s2 = Assignment(S('CL'), sympy.Integer(23))
    s3 = Assignment(S('CL'), S('CL') + S('K'))
    s4 = Assignment(S('G'), S('X'))
    s = ModelStatements([s1, s2, s3, s4])
    s.remove_symbol_definitions([S('CL'), S('K')], s4)
    assert s == ModelStatements([s4])


def test_remove_unused_parameters_and_rvs(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    model.remove_unused_parameters_and_rvs()


def test_reassign():
    s1 = Assignment(S('G'), sympy.Integer(3))
    s2 = Assignment(S('M'), sympy.Integer(2))
    s3 = Assignment(S('Z'), sympy.Integer(23) + S('M'))
    s4 = Assignment(S('KA'), S('X') + S('Y'))
    s = ModelStatements([s1, s2, s3, s4])
    s.reassign(S('M'), S('x') + S('y'))
    assert s == ModelStatements([s1, Assignment('M', S('x') + S('y')), s3, s4])

    s5 = Assignment('KA', S('KA') + S('Q') + 1)
    s = ModelStatements([s1, s2, s3, s4, s5])
    s.reassign(S('KA'), S('F'))
    assert s == ModelStatements([s1, s2, s3, Assignment('KA', S('F'))])


def test_find_compartment(testdata):
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    comp = model.statements.ode_system.find_compartment('CENTRAL')
    assert comp.name == 'CENTRAL'
    comp = model.statements.ode_system.find_compartment('NOTINMODEL')
    assert comp is None


def test_output_compartment(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    model.statements.ode_system.add_compartment("NEW")
    with pytest.raises(ValueError):
        model.statements.ode_system.output_compartment


def test_dosing_compartment(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    assert model.statements.ode_system.dosing_compartment.name == 'CENTRAL'
    model.statements.ode_system.dosing_compartment.dose = None
    with pytest.raises(ValueError):
        model.statements.ode_system.dosing_compartment


def test_central_compartment(testdata):
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.central_compartment.name == 'CENTRAL'
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan5_nodepot.mod')
    assert model.statements.ode_system.central_compartment.name == 'CENTRAL'


def test_find_depot(testdata):
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.find_depot(model.statements).name == 'DEPOT'
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan1.mod')
    assert model.statements.ode_system.find_depot(model.statements) is None
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan5_depot.mod')
    assert model.statements.ode_system.find_depot(model.statements).name == 'DEPOT'
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan5_nodepot.mod')
    assert model.statements.ode_system.find_depot(model.statements) is None


def test_peripheral_compartments(testdata):
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.peripheral_compartments == []


def test_find_transit_compartments(testdata):
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan1.mod')
    assert model.statements.ode_system.find_transit_compartments(model.statements) == []
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    assert model.statements.ode_system.find_transit_compartments(model.statements) == []
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_1transit.mod')
    transits = model.statements.ode_system.find_transit_compartments(model.statements)
    assert len(transits) == 1
    assert transits[0].name == 'TRANS1'
    model = Model(testdata / 'nonmem' / 'modeling' / 'pheno_2transits.mod')
    transits = model.statements.ode_system.find_transit_compartments(model.statements)
    assert len(transits) == 2
    assert transits[0].name == 'TRANS1'
    assert transits[1].name == 'TRANS2'


def test_insert_before_odes(testdata):
    model = Model(testdata / 'nonmem' / 'minimal.mod')
    model.statements.insert_before_odes(Assignment('CL', sympy.Integer(1)))
    assert model.model_code.split('\n')[6] == 'CL = 1'


def test_before_odes(pheno_path):
    model = Model(pheno_path)
    before_ode = model.statements.before_odes
    assert before_ode[-1].symbol.name == 'S1'


def test_full_expression(pheno_path):
    model = Model(pheno_path)
    expr = model.statements.before_odes.full_expression("CL")
    assert expr == sympy.Symbol("THETA(1)") * sympy.Symbol("WGT") * sympy.exp(
        sympy.Symbol("ETA(1)")
    )


def test_to_explicit_ode_system(pheno_path):
    model = Model(pheno_path)
    exodes = model.statements.ode_system.to_explicit_system(skip_output=True)
    odes, ics = exodes.odes, exodes.ics
    assert len(odes) == 1
    assert len(ics) == 1

    exodes = model.statements.ode_system.to_explicit_system()
    odes, ics = exodes.odes, exodes.ics
    assert len(odes) == 2
    assert len(ics) == 2


def test_repr_latex():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    latex = s1._repr_latex_()
    assert latex == r'$\displaystyle KA := \displaystyle X + Y$'


def test_repr_html():
    s1 = Assignment(S('KA'), S('X') + S('Y'))
    stats = ModelStatements([s1])
    html = stats._repr_html_()
    assert 'X + Y' in html


def test_dependencies(pheno_path):
    model = Model(pheno_path)
    depsy = model.statements.dependencies(S('Y'))
    assert depsy == {
        S('EPS(1)'),
        S('AMT'),
        S('THETA(1)'),
        S('t'),
        S('THETA(2)'),
        S('THETA(3)'),
        S('APGR'),
        S('WGT'),
        S('ETA(2)'),
        S('ETA(1)'),
    }
    depscl = model.statements.dependencies(S('CL'))
    assert depscl == {S('THETA(1)'), S('WGT'), S('ETA(1)')}



import pandas as pd
import pytest
import sympy
import sympy.physics.units as units

from pharmpy.parameter import Parameter, Parameters
from pharmpy.symbols import symbol


@pytest.mark.parametrize(
    'name,init,lower,upper,fix',
    [
        ('THETA(1)', 23, None, None, None),
        ('X', 12, None, None, None),
        ('_NAME', 0, None, None, None),
        ('OMEGA(2,1)', 0.1, 0, None, None),
        ('TCVL', 0.23, -2, 2, None),
    ],
)
def test_initialization(name, init, lower, upper, fix):
    param = Parameter(name, init, lower, upper, fix)
    assert param.name == name
    assert param.symbol == symbol(name)
    assert param.init == init
    if lower is not None:
        assert param.lower == lower
    else:
        assert param.lower == -sympy.oo
    if upper is not None:
        assert param.upper == upper
    else:
        assert param.upper == sympy.oo
    assert param.fix == bool(fix)


@pytest.mark.parametrize(
    'name,init,lower,upper,fix',
    [
        ('OMEGA(2,1)', 0.1, 2, None, None),
        ('X', 1, 0, -1, None),
    ],
)
def test_illegal_initialization(name, init, lower, upper, fix):
    with pytest.raises(ValueError):
        Parameter(name, init, lower, upper, fix)


def test_is_close_to_bound():
    param = Parameter('X', 2, lower=0, upper=23.8)
    assert param.is_close_to_bound(0.007, zero_limit=0.01, significant_digits=2) is True
    assert param.is_close_to_bound(0.007, zero_limit=0.001, significant_digits=2) is False
    assert param.is_close_to_bound(23.2, zero_limit=0.001, significant_digits=2) is False
    assert param.is_close_to_bound(23.2, zero_limit=0.001, significant_digits=1) is True
    assert param.is_close_to_bound(23.5, zero_limit=0.001, significant_digits=2) is True
    assert not param.is_close_to_bound()


def test_repr():
    param = Parameter('X', 2, lower=0, upper=23)
    assert repr(param) == 'Parameter("X", 2, lower=0, upper=23, fix=False)'


def test_copy():
    p1 = Parameter('X', 2, lower=0, upper=23)
    p2 = p1.copy()
    p1.init = 22
    assert p2.init == 2


def test_unconstrain():
    param = Parameter('X', 2, lower=0, upper=23)
    param.unconstrain()
    assert param.lower == -sympy.oo
    assert param.upper == sympy.oo

    fixed_param = Parameter('Y', 0, fix=True)
    fixed_param.unconstrain()
    assert fixed_param.lower == -sympy.oo
    assert fixed_param.upper == sympy.oo


def test_fix():
    param = Parameter('X', 2, lower=0, upper=23)
    assert param == Parameter('X', 2, lower=0, upper=23, fix=False)
    param.fix = True
    assert param == Parameter('X', 2, lower=0, upper=23, fix=True)


def test_init():
    param = Parameter('X', 2, lower=0, upper=23)
    with pytest.raises(ValueError):
        param.init = -1
    param.init = 22


def test_pset_init():
    p = Parameter('Y', 9)
    pset = Parameters([p])
    pset2 = Parameters(pset)
    assert len(pset2) == 1
    assert pset2['Y'].init == 9

    p2 = Parameter('Y', 12)
    with pytest.raises(ValueError):
        Parameters([p, p2])

    with pytest.raises(ValueError):
        Parameters([23])


def test_pset_getitem():
    p = Parameter('Y', 9)
    pset = Parameters((p,))
    assert len(pset) == 1
    assert pset['Y'] is p

    p2 = Parameter('Z', 5)
    pset.append(p2)

    assert len(pset) == 2

    # Check that the parameter set keeps the insertion order upon iteration
    for i, param in enumerate(pset):
        if i == 0:
            assert param is p
        else:
            assert param is p2

    assert pset[symbol('Z')] == p2

    p3 = Parameter('K', 19)
    with pytest.raises(KeyError):
        pset[p3]

    assert len(pset[[p]]) == 1


def test_pset_setitem():
    p1 = Parameter('P1', 1)
    p2 = Parameter('P2', 2)
    p3 = Parameter('P3', 3)
    ps = Parameters([p1, p2, p3])
    p4 = Parameter('P4', 4)
    ps[p1] = p4
    assert len(ps) == 3
    assert ps[0].name == 'P4'

    with pytest.raises(ValueError):
        ps[0] = 23

    p5 = Parameter('P4', 0)
    with pytest.raises(ValueError):
        ps[1] = p5


def test_pset_remove_fixed():
    p1 = Parameter('Y', 9, fix=False)
    p2 = Parameter('X', 3, fix=True)
    p3 = Parameter('Z', 1, fix=False)
    pset = Parameters([p1, p2, p3])
    pset.remove_fixed()
    assert len(pset) == 2
    assert pset['Y'] == Parameter('Y', 9)


def test_pset_names():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset = Parameters([p1, p2, p3])
    assert pset.names == ['Y', 'X', 'Z']
    assert pset.symbols == [symbol('Y'), symbol('X'), symbol('Z')]


def test_pset_lower_upper():
    p1 = Parameter('X', 0, lower=-1, upper=1)
    p2 = Parameter('Y', 1, lower=0)
    pset = Parameters([p1, p2])
    assert pset.lower == {'X': -1, 'Y': 0}
    assert pset.upper == {'X': 1, 'Y': sympy.oo}


def test_pset_inits():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset = Parameters([p1, p2, p3])
    pset.inits = {'X': 28}
    assert len(pset) == 3
    assert pset['X'] == Parameter('X', 28)
    assert pset['Y'] == Parameter('Y', 9)
    assert pset['Z'] == Parameter('Z', 1)

    with pytest.raises(KeyError):
        pset.inits = {'CL': 0}

    pset.inits = {'X': 0, 'Y': 2, 'Z': 5}
    assert len(pset) == 3
    assert pset['X'] == Parameter('X', 0)
    assert pset['Y'] == Parameter('Y', 2)
    assert pset['Z'] == Parameter('Z', 5)


def test_pset_nonfixed_inits():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset = Parameters([p1, p2, p3])
    assert pset.nonfixed_inits == {'Y': 9, 'X': 3, 'Z': 1}
    pset['X'].fix = True
    assert pset.nonfixed_inits == {'Y': 9, 'Z': 1}


def test_pset_fix():
    p1 = Parameter('Y', 9, fix=False)
    p2 = Parameter('X', 3, fix=True)
    p3 = Parameter('Z', 1, fix=False)
    pset = Parameters([p1, p2, p3])
    assert pset.fix == {'Y': False, 'X': True, 'Z': False}
    fixedness = {'Y': True, 'X': True, 'Z': True}
    pset.fix = fixedness
    assert pset.fix == {'Y': True, 'X': True, 'Z': True}
    with pytest.raises(KeyError):
        pset.fix = {'K': True}


def test_pset_repr():
    p1 = Parameter('Y', 9, fix=False)
    pset = Parameters([p1])
    assert type(repr(pset)) == str
    assert type(pset._repr_html_()) == str
    pset = Parameters()
    assert type(repr(pset)) == str
    assert type(pset._repr_html_()) == str


def test_parameter_space():
    p1 = Parameter('Y', 9, fix=True)
    assert p1.parameter_space == sympy.FiniteSet(9)
    p2 = Parameter('X', 10, lower=0, upper=15)
    assert p2.parameter_space == sympy.Interval(0, 15)


def test_pset_eq():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset1 = Parameters([p1, p2, p3])
    pset2 = Parameters([p1, p2])
    assert pset1 != pset2
    pset3 = Parameters([p1, p3, p2])
    assert pset1 != pset3
    assert pset1 == pset1


def test_pset_add():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset1 = Parameters([p1, p2])
    pset1.append(p3)
    assert len(pset1) == 3

    with pytest.raises(ValueError):
        pset1.append(23)


def test_pset_discard():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset1 = Parameters([p1, p2, p3])
    del pset1[p2]
    assert len(pset1) == 2
    del pset1['Y']
    assert len(pset1) == 1


def test_is_close_to_bound_pset():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3, lower=1, upper=24)
    p3 = Parameter('Z', 1, lower=0, upper=2)
    pset1 = Parameters([p1, p2, p3])
    assert not pset1.is_close_to_bound().any()
    assert not pset1.is_close_to_bound(pd.Series({'X': 3.5, 'Y': 19})).any()


def test_copy_pset():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3, lower=1, upper=24)
    p3 = Parameter('Z', 1, lower=0, upper=2)
    pset1 = Parameters([p1, p2, p3])
    pset2 = pset1.copy()
    assert pset1 == pset2
    assert id(pset1[0]) != id(pset2[0])
    p4 = p1.copy()
    assert p4 == p1
    assert id(p4) != id(p1)


def test_hash():
    p1 = Parameter('Y', 9)
    hash(p1)


def test_insert():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3, lower=1, upper=24)
    p3 = Parameter('Z', 1, lower=0, upper=2)
    pset1 = Parameters([p1, p2])
    pset1.insert(0, p3)
    assert pset1.names == ['Z', 'Y', 'X']

    p4 = Parameter('Y', 0)
    with pytest.raises(ValueError):
        pset1.insert(1, p4)


def test_simplify():
    x = sympy.Symbol('x')
    y = sympy.Symbol('y')

    p1 = Parameter('x', 3)
    p2 = Parameter('y', 9, fix=True)
    pset = Parameters([p1, p2])
    assert pset.simplify(x * y) == 9.0 * x

    p1 = Parameter('x', 3, lower=0.001)
    p2 = Parameter('y', 9)
    pset = Parameters([p1, p2])
    assert pset.simplify(abs(x)) == x

    p1 = Parameter('x', 3, lower=0)
    p2 = Parameter('y', 9)
    pset = Parameters([p1, p2])
    assert pset.simplify(sympy.Piecewise((2, sympy.Ge(x, 0)), (56, True))) == 2

    p1 = Parameter('x', -3, upper=-1)
    p2 = Parameter('y', 9)
    pset = Parameters([p1, p2])
    assert pset.simplify(abs(x)) == -x

    p1 = Parameter('x', -3, upper=0)
    p2 = Parameter('y', 9)
    pset = Parameters([p1, p2])
    assert pset.simplify(sympy.Piecewise((2, sympy.Le(x, 0)), (56, True))) == 2

    p1 = Parameter('x', 3)
    p2 = Parameter('y', 9)
    pset = Parameters([p1, p2])
    assert pset.simplify(x * y) == x * y


def test_unit():
    p = Parameter('x', 3, unit='l/h')
    assert p.unit == units.liter / units.hour
    p2 = Parameter('x', 3)
    assert p2.unit == 1



r"""
Data
====

The PharmPy data package is a standalone package.

The main class is the PharmDataFrame which is a pandas DataFrame with some extras
relevant to pharmacometrics.

Manipulation of stand alone datasets

.. list-table:: Options for the data module
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``na_values``
     - ``[-99]``
     - List
     - Data values to be converted to NA when reading in data
   * - ``na_rep``
     - ``'-99'``
     - str
     - Data value to convert NA to when writing data

"""

import pharmpy.config as config
from pharmpy.data.data_frame import ColumnType, DatasetError, DatasetWarning, PharmDataFrame
from pharmpy.data.read import read_csv, read_nonmem_dataset


class DataConfiguration(config.Configuration):
    module = 'pharmpy.data'
    na_values = config.ConfigItem(
        [-99], 'List of data values to be converted to NA when reading data'
    )
    na_rep = config.ConfigItem('-99', 'What to replace NA with in written datasets')


conf = DataConfiguration()

__all__ = [
    'DatasetError',
    'DatasetWarning',
    'ColumnType',
    'PharmDataFrame',
    'read_nonmem_dataset',
    'read_csv',
]



import os
from glob import glob
from pathlib import Path

from pharmpy.workflows import ModelExecutionEngine


class NONMEM7(ModelExecutionEngine):
    """NONMEM7 execution engine.

    An implementation of the generic :class:`~pharmpy.workflows.ModelExecutionEngine` (see for more
    information).

    Will automatically scan for installed NONMEM versions (see :func:`~self.scan_installed`).
    """

    def __init__(self, model, version=None):
        self.installed = self.scan_installed()
        if not version and self.installed:
            version = max(self.installed.keys())
        if version not in self.installed:
            raise RuntimeError(
                'Requested NONMEM7 version %r not available (found: %s)',
                version,
                ', '.join(self.installed.values()),
            )
        else:
            self.info = (version, *self.installed[version])

    def commandline(self, model):
        mod = str(self.model.source.path)
        lst = mod.with_suffix('.lst')
        if mod == lst:
            lst += '.lst'
        return (self.bin, mod, lst)

    @property
    def bin(self):
        return self.info[2] if self.info else None

    @property
    def version(self):
        return self.info[0] if self.info else None

    @staticmethod
    def scan_installed():
        if os.name == 'nt':
            globpaths = ['C:/nm7*', 'C:/nm_7*', 'C:/NONMEM/nm7*', 'C:/NONMEM/nm_7*']
            globpaths = [str(Path(globpath).resolve().expanduser()) for globpath in globpaths]
            script = 'nmfe7%s.bat'
        else:
            globpaths = []
            for root in [Path(x) for x in ('/opt', '~')]:
                tails = ['nm7*', 'nm_7*', 'nm7*', 'nm_7*', 'nonmem/nm7*', 'nonmem/nm_7*']
                globpaths += [root / x for x in tails]
            globpaths = [str(Path(globpath).expanduser()) for globpath in globpaths]
            script = 'nmfe7%s'
        candidates = [path for globpath in globpaths for path in glob(globpath)]

        nm_info = dict()
        for nm_dir in candidates:
            path = Path(nm_dir)

            if not path.is_dir():
                continue
            subdirs = ['run', 'util']
            if not any((path / subdir).is_dir() for subdir in subdirs):
                continue

            ver_suffix = {7.1: '', 7.2: '2', 7.3: '3', 7.4: '4', 7.5: '5'}
            for version, suffix in ver_suffix.items():
                nm_bin = [path / subdir / (script % suffix) for subdir in subdirs]
                nm_bin = list(filter(lambda x: x.is_file(), nm_bin))
                if nm_bin:
                    nm_info[version] = (nm_dir, str(nm_bin[0]))
                    break

        return nm_info



from pathlib import Path

import pharmpy.utils as utils


def test_tempdir(fs):
    with utils.TemporaryDirectory() as td:
        path = Path(td)
        assert path.is_dir()
    assert not path.is_dir()



from pharmpy import Model
from pharmpy.modeling import load_example_model, plot_individual_predictions, plot_iofv_vs_iofv


def test_plot_iofv_vs_iofv():
    model = load_example_model('pheno')
    assert plot_iofv_vs_iofv(model, model)


def test_plot_individual_predictions(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    plot = plot_individual_predictions(model)
    assert plot
    plot = plot_individual_predictions(model, predictions=['PRED'], individuals=[1, 2, 5])
    assert plot



import re

from pharmpy.model import ModelSyntaxError

from .records.factory import create_record
from .records.raw_record import RawRecord


class Abbreviated:
    """Handling all $ABBREVIATED in a control stream"""

    def __init__(self, stream):
        self.stream = stream

    @property
    def replace(self):
        """Get all $ABBR REPLACE as a dictionary"""
        d = dict()
        for record in self.stream.get_records('ABBREVIATED'):
            d.update(record.replace)
        return d

    def translate_to_pharmpy_names(self):
        d = dict()
        for record in self.stream.get_records('ABBREVIATED'):
            d.update(record.translate_to_pharmpy_names())
        return d


class NMTranParser:
    """Parser for NMTran control streams"""

    def parse(self, text):
        stream = NMTranControlStream()

        record_strings = re.split(r'^([ \t]*\$)', text, flags=re.MULTILINE)
        first = record_strings.pop(0)  # Empty if nothing before first record
        if first:
            stream.records.append(RawRecord(first))

        for separator, s in zip(record_strings[0::2], record_strings[1::2]):
            record = create_record(separator + s)
            stream.records.append(record)

        return stream


default_record_order = [
    'SIZES',
    'SUBROUTINES',
    'MODEL',
    'ABBREVIATED',
    'PK',
    'PRED',
    'DES',
    'ERROR',
    'THETA',
    'OMEGA',
    'SIGMA',
    'ESTIMATION',
    'TABLE',
]


class NMTranControlStream:
    """Representation of a parsed control stream (model file)"""

    def __init__(self):
        self.records = []
        self._active_problem = 0
        self.abbreviated = Abbreviated(self)

    def get_records(self, name):
        """Return a list of all records of a certain type in the current $PROBLEM"""
        current_problem = -1
        found = []
        for record in self.records:
            if record.name == 'PROBLEM':
                current_problem += 1
            if current_problem == self._active_problem and record.name == name:
                found.append(record)
        return found

    def append_record(self, content):
        """Create and append record at the end"""
        record = create_record(content)
        self.records.append(record)
        return record

    def insert_record(self, content, at_index=None):
        """Create and insert a new record at the correct position

        If the record type is already present the new record will be put
        directly after the last record of that type.
        If no record of the type is present the new record will be put
        given the default record order.
        If record type is unknown. Place at the end.
        """

        record = create_record(content)
        name = record.name

        if at_index:
            self.records.insert(at_index, record)
            return record

        current_problem = -1
        index = None
        for i, currec in enumerate(self.records):
            if currec.name == 'PROBLEM':
                current_problem += 1
            if current_problem == self._active_problem and currec.name == name:
                index = i

        if index is None:
            try:
                default_order_index = default_record_order.index(name)
                before_records = default_record_order[0:default_order_index]
            except ValueError:
                before_records = []
            current_problem = -1
            for i, currec in enumerate(self.records):
                if currec.name == 'PROBLEM':
                    current_problem += 1
                if current_problem == self._active_problem and currec.name in before_records:
                    index = i

        if index is None:
            index = len(self.records)

        self.records.insert(index + 1, record)
        return record

    def remove_records(self, records):
        for rec in records:
            self.records.remove(rec)

    def replace_records(self, old, new):
        keep = []
        first = True
        for rec in self.records:
            if rec not in old:
                keep.append(rec)
            else:
                if first:
                    keep.extend(new)
                    first = False
        self.records = keep

    def go_through_omega_rec(self):
        next_omega = 1
        previous_size = None
        prev_cov = None

        for omega_record in self.get_records('OMEGA'):
            next_omega_cur = next_omega
            omegas, next_omega, previous_size = omega_record.parameters(
                next_omega_cur, previous_size
            )
            etas, next_eta, prev_cov, _ = omega_record.random_variables(next_omega_cur, prev_cov)

    def validate(self):
        in_problem = False
        for record in self.records:
            if in_problem and record.name == 'SIZES':
                raise ModelSyntaxError('The SIZES record must come before the first PROBLEM record')
            elif record.name == 'PROBLEM':
                in_problem = True
            if hasattr(record, 'validate'):
                record.validate()

    def __str__(self):
        return ''.join(str(x) for x in self.records)



from pathlib import Path

import pandas as pd

from pharmpy import Model
from pharmpy.results import Results


class LinearizeResults(Results):
    def __init__(self, ofv=None, iofv=None):
        self.ofv = ofv
        self.iofv = iofv


def calculate_results(base_model, linear_model):
    baseres = base_model.modelfit_results
    linearres = linear_model.modelfit_results
    linearres.evaluation_ofv
    iofv = pd.DataFrame(
        {
            'base': baseres.individual_ofv,
            'linear': linearres.individual_ofv,
            'delta': linearres.individual_ofv - baseres.individual_ofv,
        }
    )
    ofv = pd.DataFrame(
        {'ofv': [baseres.ofv, linearres.evaluation_ofv, linearres.ofv]},
        index=['base', 'lin_evaluated', 'lin_estimated'],
    )
    res = LinearizeResults(ofv=ofv, iofv=iofv)
    return res


def psn_linearize_results(path):
    """Create linearize results from a PsN linearize run

    :param path: Path to PsN linearize run directory
    :return: A :class:`QAResults` object
    """
    path = Path(path)

    base_model = Model(path / 'scm_dir1' / 'derivatives.mod')
    lin_path = list(path.glob('*_linbase.mod'))[0]
    lin_model = Model(lin_path)

    res = calculate_results(base_model, lin_model)
    return res



import re
from io import StringIO
from pathlib import Path

import numpy as np
import pandas as pd

import pharmpy.math


class NONMEMTableFile:
    """A NONMEM table file that can contain multiple tables"""

    def __init__(self, path=None, tables=None, notitle=False, nolabel=False):
        if path is not None:
            path = Path(path)
            suffix = path.suffix
            self.tables = []
            with open(str(path), 'r') as tablefile:
                if notitle:
                    self._add_table(tablefile.read().splitlines(), notitle=notitle, nolabel=nolabel)
                else:
                    current = []
                    for line in tablefile:
                        if line.startswith("TABLE NO."):
                            if current:
                                self._add_table(current, suffix)
                            current = [line]
                        else:
                            current.append(line)
                    self._add_table(current, suffix)
            self._count = 0
        else:
            self.tables = tables

    def __iter__(self):
        return self

    def _add_table(self, content, suffix=None, notitle=False, nolabel=False):
        if not notitle:
            table_line = content.pop(0)
        if suffix == '.ext':
            table = ExtTable(''.join(content))
        elif suffix == '.phi':
            table = PhiTable(''.join(content))
        elif suffix == '.cov' or suffix == '.cor' or suffix == '.coi':
            table = CovTable(''.join(content))
        else:
            table = NONMEMTable(''.join(content))  # Fallback to non-specific table type
        if not notitle:
            m = re.match(r'TABLE NO.\s+(\d+)', table_line)
            if not m:
                raise ValueError(f"Illegal {suffix}-file: missing TABLE NO.")
            table.number = int(m.group(1))
            table.is_evaluation = False
            if re.search(r'(Evaluation)', table_line):
                table.is_evaluation = True  # No estimation step was run
            m = re.match(
                r'TABLE NO.\s+\d+: (.*?): (?:Goal Function=(.*): )?Problem=(\d+) '
                r'Subproblem=(\d+) Superproblem1=(\d+) Iteration1=(\d+) Superproblem2=(\d+) '
                r'Iteration2=(\d+)',
                table_line,
            )
            if m:
                table.method = m.group(1)
                table.goal_function = m.group(2)
                table.problem = int(m.group(3))
                table.subproblem = int(m.group(4))
                table.superproblem1 = int(m.group(5))
                table.iteration1 = int(m.group(6))
                table.superproblem2 = int(m.group(7))
                table.iteration2 = int(m.group(8))
        self.tables.append(table)

    def __len__(self):
        return len(self.tables)

    @property
    def table(
        self, problem=1, subproblem=0, superproblem1=0, iteration1=0, superproblem2=0, iteration2=0
    ):
        for t in self.tables:
            if (
                t.problem == problem
                and t.subproblem == subproblem
                and t.superproblem1 == superproblem1
                and t.iteration1 == iteration1
                and t.superproblem2 == superproblem2
                and t.iteration2 == iteration2
            ):
                return t

    def table_no(self, table_number=1):
        for table in self.tables:
            if table.number == table_number:
                return table

    def write(self, path):
        with open(path, 'w') as df:
            for table in self.tables:
                print('TABLE NO.     1', file=df)
                table.create_content()
                print(table.content, file=df, end='')

    def __next__(self):
        if self._count >= len(self):
            raise StopIteration
        else:
            self._count += 1
            return self.tables[self._count - 1]


class NONMEMTable:
    """A NONMEM output table."""

    def __init__(self, content=None, df=None):
        if content is not None:
            self._df = pd.read_table(StringIO(content), sep=r'\s+', engine='python')
        else:
            self._df = df

    @property
    def data_frame(self):
        return self._df

    @staticmethod
    def rename_index(df, ext=True):
        """If columns rename also the row index"""
        theta_labels = [x for x in df.columns if x.startswith('THETA')]
        omega_labels = [x for x in df.columns if x.startswith('OMEGA')]
        sigma_labels = [x for x in df.columns if x.startswith('SIGMA')]
        labels = theta_labels + omega_labels + sigma_labels
        if ext:
            labels = ['ITERATION'] + labels + ['OBJ']
        else:
            df = df.reindex(labels)
        df = df.reindex(labels, axis=1)
        df.columns = df.columns.str.replace(r'THETA(\d+)', r'THETA(\1)', regex=True)
        if not ext:
            df.index = df.index.str.replace(r'THETA(\d+)', r'THETA(\1)', regex=True)
        return df


class CovTable(NONMEMTable):
    @property
    def data_frame(self):
        df = self._df.copy(deep=True)
        df.set_index('NAME', inplace=True)
        df.index.name = None
        df = NONMEMTable.rename_index(df, ext=False)
        df = df.loc[(df != 0).any(axis=1), (df != 0).any(axis=0)]  # Remove FIX
        return df


class PhiTable(NONMEMTable):
    @property
    def iofv(self):
        df = self._df.copy(deep=True)
        # Remove all zero lines. These are from individuals with no observations
        df = df.loc[df.iloc[:, 2:].any(axis=1)]
        iofv = df[['ID', 'OBJ']]
        iofv.set_index('ID', inplace=True)
        iofv.columns = ['iOFV']
        return iofv['iOFV']

    @property
    def etas(self):
        df = self._df.copy(deep=True)
        df = df.loc[df.iloc[:, 2:].any(axis=1)]
        eta_col_names = [col for col in df if col.startswith('ETA')]
        etas = df[eta_col_names]
        etas.index = df['ID']
        return etas

    @property
    def etcs(self):
        df = self._df.copy(deep=True)
        df = df.loc[df.iloc[:, 2:].any(axis=1)]
        eta_col_names = [col for col in df if col.startswith('ETA')]
        etc_col_names = [col for col in df if col.startswith('ETC')]
        vals = df[etc_col_names].values
        matrix_array = [pharmpy.math.flattened_to_symmetric(x) for x in vals]
        etc_frames = [
            pd.DataFrame(matrix, columns=eta_col_names, index=eta_col_names)
            for matrix in matrix_array
        ]
        etcs = pd.Series(etc_frames, index=df['ID'])
        return etcs

    def create_content(self):
        df = self._df.copy(deep=True)
        df.reset_index(inplace=True)
        df.insert(loc=0, column='SUBJECT_NO', value=np.arange(len(df)) + 1)
        fmt = '%13d%13d' + '%13.5E' * (len(df.columns) - 2)
        with StringIO() as s:
            np.savetxt(s, df.values, fmt=fmt)
            self.content = s.getvalue()
        with StringIO() as s:
            header_fmt = ' %-12s' * len(df.columns) + '\n'
            header = header_fmt % tuple(df.columns)
            self.content = header + self.content


class ExtTable(NONMEMTable):
    @property
    def data_frame(self):
        df = self._df.copy(deep=True)
        df = NONMEMTable.rename_index(df)
        if df['ITERATION'].isnull().values.all():
            raise ValueError('Broken table in ext-file')
        return df

    def _get_parameters(self, iteration, include_thetas=True):
        df = self.data_frame
        row = df.loc[df['ITERATION'] == iteration]
        if df['ITERATION'].isnull().values.all():
            return np.nan
        elif row.empty:
            raise KeyError(f'Row {iteration} not available in ext-file')
        del row['ITERATION']
        del row['OBJ']
        if not include_thetas:
            row = row[row.columns.drop(list(row.filter(regex='THETA')))]
        return row.squeeze()

    def _get_ofv(self, iteration):
        df = self.data_frame
        row = df.loc[df['ITERATION'] == iteration]
        if df['ITERATION'].isnull().values.all():
            return np.nan
        elif row.empty:
            raise KeyError(f'Row {iteration} not available in ext-file')
        row.squeeze()
        return row['OBJ'].squeeze()

    @property
    def iterations(self):
        """List of all iteration numbers (except the special negative)"""
        it = self.data_frame['ITERATION']
        return list(it[it >= 0])

    @property
    def final_parameter_estimates(self):
        """Get the final parameter estimates as a Series
        A specific parameter estimate can be retreived as
        final_parameter_estimates['THETA1']
        Fallback to estimates from last iteration in case
        of an aborted or failed run.
        """
        try:
            ser = self._get_parameters(-1000000000)
        except KeyError:
            final_iter = max(self.iterations)
            ser = self._get_parameters(final_iter)
        ser.name = 'estimates'
        return ser

    @property
    def standard_errors(self):
        """Get the standard errors of the parameter estimates as a Series"""
        ser = self._get_parameters(-1000000001)
        ser.name = 'SE'
        return ser

    @property
    def condition_number(self):
        """Get the condition number of the correlation matrix"""
        ser = self._get_parameters(-1000000003)
        return ser.values[0]

    @property
    def omega_sigma_stdcorr(self):
        """Get the omegas and sigmas in standard deviation/correlation form"""
        return self._get_parameters(-1000000004, include_thetas=False)

    @property
    def omega_sigma_se_stdcorr(self):
        """The standard error of the omegas and sigmas in stdcorr form"""
        return self._get_parameters(-1000000005, include_thetas=False)

    @property
    def fixed(self):
        """What parameters were fixed?"""
        fix = self._get_parameters(-1000000006)
        return fix.apply(bool)

    @property
    def final_ofv(self):
        try:
            ser = self._get_ofv(-1000000000)
        except KeyError:
            final_iter = max(self.iterations)
            ser = self._get_ofv(final_iter)
        return ser

    @property
    def initial_ofv(self):
        try:
            ser = self._get_ofv(0)
        except KeyError:
            ser = self._get_ofv(-1000000000)
        return ser



"""
:meta private:
"""

import sympy

from pharmpy.modeling.help_functions import _format_input_list, _get_epsilons
from pharmpy.parameter import Parameter
from pharmpy.random_variables import RandomVariable
from pharmpy.symbols import symbol as S


def set_iiv_on_ruv(model, list_of_eps=None, same_eta=True, eta_names=None):
    """
    Multiplies epsilons with exponential (new) etas.

    Initial variance for new etas is 0.09.

    Parameters
    ----------
    model : Model
        Pharmpy model to apply IIV on epsilons.
    list_of_eps : str, list
        Name/names of epsilons to multiply with exponential etas. If None, all epsilons will
        be chosen. None is default.
    same_eta : bool
        Boolean of whether all RUVs from input should use the same new ETA or if one ETA
        should be created for each RUV. True is default.
    eta_names : str, list
        Custom names of new etas. Must be equal to the number epsilons or 1 if same eta.

    Return
    ------
    Model
        Reference to same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_iiv_on_ruv(model)   # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                   ETA_RV1
    Y := EPS(1)⋅W⋅ℯ        + F

    See also
    --------
    set_power_on_ruv

    """
    list_of_eps = _format_input_list(list_of_eps)
    eps = _get_epsilons(model, list_of_eps)

    if eta_names and len(eta_names) != len(eps):
        raise ValueError(
            'The number of provided eta names must be equal to the number of epsilons.'
        )

    rvs, pset, sset = model.random_variables, model.parameters, model.statements

    if same_eta:
        eta = _create_eta(pset, 1, eta_names)
        rvs.append(eta)
        eta_dict = {e: eta for e in eps}
    else:
        etas = [_create_eta(pset, i + 1, eta_names) for i in range(len(eps))]
        for eta in etas:
            rvs.append(eta)
        eta_dict = {e: eta for e, eta in zip(eps, etas)}

    for e in eps:
        sset.subs({e.symbol: e.symbol * sympy.exp(S(eta_dict[e].name))})

    model.random_variables = rvs
    model.parameters = pset
    model.statements = sset

    model.modelfit_results = None

    return model


def _create_eta(pset, number, eta_names):
    omega = S(f'IIV_RUV{number}')
    pset.append(Parameter(str(omega), 0.09))

    if eta_names:
        eta_name = eta_names[number - 1]
    else:
        eta_name = f'ETA_RV{number}'

    eta = RandomVariable.normal(eta_name, 'iiv', 0, omega)
    return eta



from pathlib import Path

import pandas as pd

from pharmpy.results import Results


class ResmodResults(Results):
    """Resmod results class"""

    def __init__(self, models=None):
        self.models = models


def calculate_results(base_model, iiv_on_ruv, power):
    base_ofv = base_model.modelfit_results.ofv
    dofv_iiv_on_ruv = iiv_on_ruv.modelfit_results.ofv - base_ofv
    dofv_power = power.modelfit_results.ofv - base_ofv
    omega_iiv_on_ruv = iiv_on_ruv.modelfit_results.parameter_estimates["IIV_RUV1"]
    theta_power = power.modelfit_results.parameter_estimates["power1"]
    df = pd.DataFrame(
        {
            'model': ['IIV_on_RUV', 'power'],
            'dvid': 1,
            'iteration': 1,
            'dofv': [dofv_iiv_on_ruv, dofv_power],
            'parameters': [{'omega': omega_iiv_on_ruv}, {'theta': theta_power}],
        }
    )
    df.set_index(['model', 'dvid', 'iteration'], inplace=True)
    res = ResmodResults(models=df)
    return res


def psn_resmod_results(path):
    path = Path(path)
    res = ResmodResults()
    respath = path / 'resmod_results.csv'
    if respath.is_file():
        df = pd.read_csv(respath, names=range(40), skiprows=[0], engine='python')
    df[0].fillna(1, inplace=True)
    df[1].fillna(1, inplace=True)
    df.dropna(how='all', axis=1, inplace=True)
    df2 = df[[0, 1, 2, 3]].copy()
    df2 = df2.astype({0: int})
    df2.columns = ['iteration', 'DVID', 'model', 'dOFV']
    df2.set_index(['iteration', 'DVID', 'model'], inplace=True)
    parameters = pd.Series(name='parameters', index=df.index, dtype=object)
    for rowind, row in df.iterrows():
        d = dict()
        for i in range(4, len(row)):
            if row[i] is not None:
                a = row[i].split('=')
                d[a[0]] = float(a[1])
        parameters[rowind] = d
    parameters.index = df2.index
    df2['parameters'] = parameters
    res.models = df2
    return res



from pharmpy.modeling import load_example_model, predict_outliers


def test_predict_outliers():
    model = load_example_model('pheno')
    res = predict_outliers(model)
    assert len(res) == 59
    assert res['residual'][1] == -0.27215176820755005



import pharmpy.tools
from pharmpy.data.iterators import Resample
from pharmpy.tools.bootstrap.results import calculate_results
from pharmpy.tools.modelfit import create_multiple_fit_workflow
from pharmpy.workflows import Task, Workflow


class Bootstrap(pharmpy.tools.Tool):
    def __init__(self, model, resamples=1):
        self.model = model
        self.resamples = resamples
        super().__init__()
        self.model.database = self.database.model_database

    def run(self):
        wf_bootstrap = self.create_workflow()
        res = self.dispatcher.run(wf_bootstrap, self.database)
        res.to_json(path=self.database.path / 'results.json')
        res.to_csv(path=self.database.path / 'results.csv')
        return res

    def create_workflow(self):
        wf = Workflow()

        for i in range(self.resamples):
            task_resample = Task('resample', resample_model, self.model, f'bs_{i + 1}')
            wf.add_task(task_resample)

        wf_fit = create_multiple_fit_workflow(n=self.resamples)
        wf.insert_workflow(wf_fit)

        task_result = Task('results', post_process_results, self.model)
        wf.add_task(task_result, predecessors=wf.output_tasks)

        return wf


def resample_model(model, name):
    resample = Resample(model, model.dataset.pharmpy.id_label, resamples=1, name=name)
    model, _ = next(resample)
    return model


def post_process_results(original_model, *models):
    res = calculate_results(
        models, original_model=original_model, included_individuals=None, dofv_results=None
    )
    return res
